/*
 * Copyright (C) 2004-2005 Claudio "KLaN" Ciccani <klan@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "df_xine.h"


#ifdef CLAMP
# undef CLAMP
#endif
#define CLAMP( n, min, max )  ((n) < (min)) ? (min) : (((n) > (max)) ? (max) : (n))

#define DELAY_LIMIT      500   /* in milliseconds */
#define STEP_SEEK      15000   /* in milliseconds */
#define STEP_COLORS      514
#define STEP_VOLUME        4
#define OSD_TIMEOUT     3500   /* in milliseconds */

/* Control Flags */
#define DCF_SET   0x00000001
#define DCF_SHOW  0x00000002
#define DCF_ALL   (DCF_SET|DCF_SHOW)



static inline unsigned long
millisec( const struct timeval *t )
{
     unsigned long msec;

     if (!t) {
          struct timeval tv;
          gettimeofday( &tv, NULL );
          msec = (tv.tv_sec*1000ul + tv.tv_usec/1000ul);
     } else
          msec = (t->tv_sec*1000ul + t->tv_usec/1000ul);

     return msec;
}

          
static const char*
dfx_error( DFXCore *this )
{
     int err = xine_get_error( this->stream );

     switch (err) {
          case XINE_ERROR_NO_INPUT_PLUGIN:
               return "no input plugin";

          case XINE_ERROR_NO_DEMUX_PLUGIN:
               return "no demuxer plugin";

          case XINE_ERROR_DEMUX_FAILED:
               return "demux failed";

          case XINE_ERROR_MALFORMED_MRL:
               return "malformed mrl";

          default:
               break;
     }

     return "generic error";
}

static bool
dfx_getpl( DFXCore *this )
{
     int pos = -1;
     int len = -1;
     int try =  5;

     while (try--) {
          if (xine_get_pos_length( this->stream,
                                   NULL, &pos, &len ))
               break;

          usleep( 1000 );
     }

     if (pos < 0 || len < 1)
          return false;

     this->ctl.pos = pos;
     this->ctl.len = len;

     return true;
}


static bool
dfx_control_play( DFXCore *this, int toggle )
{
     DFXVideoControl *ctl    = &this->ctl;
     DFXMedia        *media  = &this->media;
     
     if (toggle) {
          if (xine_get_status( this->stream ) == XINE_STATUS_PLAY) {
               DBUG( "stopping mrl '%s'\n", media->cur->mrl );
               ctl->status = XINE_STATUS_STOP;
               dfx_display_status( this, DDF_INFO );
               xine_stop( this->stream );
               usleep( 100 );
               return true;
          }
     }
     else {
          int width, height;
          
          xine_stop( this->stream );
          xine_close( this->stream );

          if (!media->cur) {
               SAY( "playlist is empty\n" );
               return false;
          }
          
          if (!xine_open( this->stream, media->cur->mrl )) {
               SAY( "skipping mrl '%s' [%s]\n",
                    media->cur->mrl, dfx_error( this ) );
               dfx_playlist_discard( this );
               return dfx_control_play( this, 0 );
          }

          width  = xine_get_stream_info( this->stream, 
                                         XINE_STREAM_INFO_VIDEO_WIDTH );
          height = xine_get_stream_info( this->stream,
                                         XINE_STREAM_INFO_VIDEO_HEIGHT );
          if (width > 1 && height > 1)
               dfx_context_configure( this, width, height, DSPF_UNKNOWN );
     }
          
     if (!xine_get_stream_info( this->stream,
                                XINE_STREAM_INFO_HAS_VIDEO ))
          dfx_context_set_gmode( this, DGM_ANIM );
     else
          dfx_context_set_gmode( this, DGM_VIDEO );

     if (!xine_play( this->stream, 0, 0 )) {
          SAY( "skipping mrl '%s' [%s]\n",
               media->cur->mrl, dfx_error( this ) );
          dfx_playlist_discard( this );
          return dfx_control_play( this, 0 );
     }
     
     SAY( "playing mrl '%s'\n", media->cur->mrl );

     usleep( 200 );
     
     return true;
}

static void
dfx_control_pause( DFXCore *this, int flags )
{
     DFXVideoControl *ctl   = &this->ctl;

     if (flags & DCF_SET) {
          if (xine_get_status( this->stream ) == XINE_STATUS_PLAY) {
               int speed = xine_get_param( this->stream, XINE_PARAM_SPEED );
               
               xine_set_param( this->stream,
                               XINE_PARAM_SPEED,
                               (speed != XINE_SPEED_PAUSE)
                                ? XINE_SPEED_PAUSE
                                : XINE_SPEED_NORMAL );
               usleep( 100 );
          }
     }

     if (flags & DCF_SHOW) {
          if (dfx_getpl( this )) {
               ctl->speed = xine_get_param( this->stream,
                                            XINE_PARAM_SPEED );
               dfx_display_status( this, DDF_INFO );
          }
     }
}

static void
dfx_control_seek( DFXCore *this, int flags, int param )
{
     DFXVideoControl *ctl = &this->ctl;
     
     if (!xine_get_stream_info( this->stream,
                                XINE_STREAM_INFO_SEEKABLE ))
          return;
     
     if (flags & DCF_SET) {
          if (xine_get_status( this->stream ) == XINE_STATUS_PLAY) {
               if (!dfx_getpl( this ))
                    return;

               switch (param) {
                    case DAP_MINUS:
                    case DAP_PLUS:
                         ctl->pos += param * STEP_SEEK; /* seek relative */
                         break;

                    default:
                         ctl->pos  = (ctl->len * param) / 100; /* seek absolute */
                         break;
               }
               ctl->pos = CLAMP( ctl->pos, 0, ctl->len );
          
               xine_play( this->stream, 0, ctl->pos );
               usleep( 100 );
          }
     }

     if (flags & DCF_SHOW) {
          if (dfx_getpl( this ))
               dfx_display_status( this, DDF_BAR );
     }
}

static void
dfx_control_speed( DFXCore *this, int flags, int param )
{
     DFXVideoControl *ctl   = &this->ctl;
     
     if (flags & DCF_SET) {
          int speed = xine_get_param( this->stream, 
                                      XINE_PARAM_SPEED );
          
          if (speed != XINE_SPEED_PAUSE) {
               if (param < 0)
                    speed = (speed + 1) >> 1;
               else
                    speed = (speed < XINE_SPEED_FAST_4) ?
                            (speed << 1) : speed;

               xine_set_param( this->stream,
                               XINE_PARAM_SPEED,
                               speed );
          }
     }

     if (flags & DCF_SHOW) {
          if (dfx_getpl( this )) {
               ctl->speed = xine_get_param( this->stream,
                                            XINE_PARAM_SPEED );
               dfx_display_status( this, DDF_INFO );
          }
     }
}

static struct {
     const char *name;
     int         cap;
     int         flag;
     int         param;
} level_data[] = {
     { "Brightness", DLCAPS_BRIGHTNESS, DCAF_BRIGHTNESS, XINE_PARAM_VO_BRIGHTNESS },
     { "Contrast",   DLCAPS_CONTRAST,   DCAF_CONTRAST,   XINE_PARAM_VO_CONTRAST   },
     { "Saturation", DLCAPS_SATURATION, DCAF_SATURATION, XINE_PARAM_VO_SATURATION },
     { "Hue",        DLCAPS_HUE,        DCAF_HUE,        XINE_PARAM_VO_HUE        }
};

static void
dfx_control_correction( DFXCore *this, int level, int flags, int param )
{
     DFXVideoControl    *ctl     = &this->ctl;
     DFXVideoContext    *ctx     = &this->ctx;
     DFBColorAdjustment *reset   = &ctl->reset;
     DFBColorAdjustment  current = { 0, 0, 0, 0, 0 };
     int                 value   = 0;
     
     if (flags & DCF_SET) {
          if (ctx->caps & level_data[level-DAI_BRIGHTNESS].cap) {
               ctx->layer->GetColorAdjustment( ctx->layer, &current );
          
               switch (level) {
                    case DAI_BRIGHTNESS:
                         value = (param == DAP_RESET) 
                                  ? reset->brightness 
                                  : current.brightness;
                         break;
                    case DAI_CONTRAST:
                         value = (param == DAP_RESET)
                                 ? reset->contrast
                                 : current.contrast;
                         break;
                    case DAI_SATURATION:
                         value = (param == DAP_RESET)
                                 ? reset->saturation
                                 : current.saturation;
                         break;
                    case DAI_HUE:
                         value = (param == DAP_RESET)
                                 ? reset->hue
                                 : current.hue;
                         break;
                    default:
                         return;
               }
          } else {
               if (param != DAP_RESET)
                    value = xine_get_param( this->stream,
                                       level_data[level-DAI_BRIGHTNESS].param );
               else
                    value = 0x8000;
          }

          if (param != DAP_RESET) {
               value += param * STEP_COLORS;
               value  = CLAMP( value, 0, 0xffff );
          }
          
          if (ctx->caps & level_data[level-DAI_BRIGHTNESS].cap) {
               current.flags      = level_data[level-DAI_BRIGHTNESS].flag;
               current.brightness = value;
               current.contrast   = value;
               current.saturation = value;
               current.hue        = value;
               
               ctx->layer->SetColorAdjustment( ctx->layer, &current );
          } else
               xine_set_param( this->stream,
                               level_data[level-DAI_BRIGHTNESS].param,
                               value );
     }

     if (flags & DCF_SHOW) {
          if (ctx->caps & level_data[level-DAI_BRIGHTNESS].cap) {
               ctx->layer->GetColorAdjustment( ctx->layer, &current );
          
               switch (level) {
                    case DAI_BRIGHTNESS:
                         value = current.brightness;
                         break;
                    case DAI_CONTRAST:
                         value = current.contrast;
                         break;
                    case DAI_SATURATION:
                         value = current.saturation;
                         break;
                    case DAI_HUE:
                         value = current.hue;
                         break;
                    default:
                         return;
               }
          } else
               value = xine_get_param( this->stream,
                                       level_data[level-DAI_BRIGHTNESS].param );
          
          dfx_display_bar( this,
                           level_data[level-DAI_BRIGHTNESS].name,
                           ((value+257) * 100) / 0xffff );
     }
}

static void
dfx_control_ratio( DFXCore *this, int flags, int param )
{
     const char *ratio_name[] = {
          [XINE_VO_ASPECT_AUTO]       = "auto",
          [XINE_VO_ASPECT_SQUARE]     = "Square",
          [XINE_VO_ASPECT_4_3]        = "4:3",
          [XINE_VO_ASPECT_ANAMORPHIC] = "Anamorphic",
          [XINE_VO_ASPECT_DVB]        = "DVB"
     };

     if (flags & DCF_SET) {
          if (param == DAP_RATIO_SCREEN) {
               /* Screen aspect ratio */
               param = (this->ctx.screen.width << 16) |
                       (this->ctx.screen.height & 0xffff);
          }
               
          xine_set_param( this->stream,
                          XINE_PARAM_VO_ASPECT_RATIO,
                          param );
     }

     if (flags & DCF_SHOW) {
          int ratio;
          
          ratio = xine_get_param( this->stream, 
                                   XINE_PARAM_VO_ASPECT_RATIO );
          if (ratio > XINE_VO_ASPECT_DVB) {
               dfx_display_text( this, "Ratio: %d:%d", 
                                       ratio>>16, ratio&0xffff );
          } else {
               dfx_display_text( this, "Ratio: %s", ratio_name[ratio] );
          }
     }
}

static void
dfx_control_zoom( DFXCore *this, int flags, int param )
{
     DFXVideoContext *ctx = &this->ctx;
     
     if (flags & DCF_SET) {
          switch (param) {
               case DAP_PLUS:
                    ctx->video.zoom += 0.01;
                    break;
               case DAP_MINUS:
                    if (ctx->video.zoom >= 0.01)
                         ctx->video.zoom -= 0.01;
                    break;
               case DAP_RESET:
                    ctx->video.zoom  = 1.0;
                    break;
          }
          
          ctx->update = true;
     }

     if (flags & DCF_SHOW)
          dfx_display_text( this, "Zoom: %d%%", (int)(ctx->video.zoom * 100.0) );
}

static void 
dfx_control_volume( DFXCore *this, int flags, int param )
{
     int vol;
     
     if (flags & DCF_SET) {
          vol = xine_get_param( this->stream, 
                                XINE_PARAM_AUDIO_VOLUME );
         
          vol += param * STEP_VOLUME;
          vol  = CLAMP( vol, 0, 100 );
          
          xine_set_param( this->stream,
                          XINE_PARAM_AUDIO_VOLUME,
                          vol );
     }

     if (flags & DCF_SHOW) {
          vol = xine_get_param( this->stream,
                                XINE_PARAM_AUDIO_VOLUME );
          dfx_display_bar( this, "Volume", vol );
     }
}          

static void
dfx_control_audio( DFXCore *this, int flags, int param )
{
     int channel;

     if (flags & DCF_SET) {
          channel = xine_get_param( this->stream,
                                    XINE_PARAM_AUDIO_CHANNEL_LOGICAL );

          if (param == DAP_RESET)
               channel  = -1;
          else
               channel += param;

          xine_set_param( this->stream,
                          XINE_PARAM_AUDIO_CHANNEL_LOGICAL,
                          channel );
          usleep( 100 );
     }
     
     if (flags & DCF_SHOW) {
          char buffer[32];
          
          channel = xine_get_param( this->stream,
                                    XINE_PARAM_AUDIO_CHANNEL_LOGICAL );

          if (!xine_get_audio_lang( this->stream, channel, &buffer[0] ))
               sprintf( buffer, "auto" );

          dfx_display_text( this, "Audio Channel: %s", buffer );
     }
}

static void
dfx_control_spu( DFXCore *this, int flags, int param )
{
     int channel;

     if (flags & DCF_SET) {
          channel = xine_get_param( this->stream,
                                    XINE_PARAM_SPU_CHANNEL );

          if (param == DAP_RESET)
               channel  = -1;
          else
               channel += param;

          if (channel < -1)
               channel = -1; /* avoid turning off spu (will disable OSD) */

          xine_set_param( this->stream,
                          XINE_PARAM_SPU_CHANNEL,
                          channel );
     }
     
     if (flags & DCF_SHOW) {
          char buffer[32];
     
          channel = xine_get_param( this->stream,
                                    XINE_PARAM_SPU_CHANNEL );

          if (!xine_get_spu_lang( this->stream, channel, &buffer[0] ))
               sprintf( buffer, "auto" );

          dfx_display_text( this, "Subtitles Channel: %s", buffer );
     }
}

static void
dfx_control_deinterlace( DFXCore *this, int flags, int param )
{
     DFXVideoControl *ctl = &this->ctl;

     if (flags & DCF_SET) { 
          IDirectFBDisplayLayer *layer   = this->ctx.layer;
          IDirectFBSurface      *surface = this->ctx.surface;
          DFBDisplayLayerConfig  config  = this->ctx.config;
          DFBResult              err;
          
          switch (param) {
               case DAP_FIELD_AUTO:
                    ctl->deinterlace = false;
                    break;
               
               case DAP_FIELD_TOP:
               case DAP_FIELD_BOTTOM:
                    ctl->deinterlace = true;
                    ctl->field       = param;
                    break;

               default:
                    return;
          }
     
          if (ctl->deinterlace) {
               if (this->ctx.caps & DLCAPS_DEINTERLACING) {
                    config.flags    = DLCONF_OPTIONS;
                    config.options |= DLOP_DEINTERLACING;

                    err = layer->SetConfiguration( layer, &config );
                    
                    if (err == DFB_OK) {
                         surface->SetField( surface, ctl->field - 1 );
                         xine_set_param( this->stream,
                                         XINE_PARAM_VO_DEINTERLACE,
                                         0 );
                    } else
                         xine_set_param( this->stream,
                                         XINE_PARAM_VO_DEINTERLACE,
                                         ctl->field );
               }
               else {
                    xine_set_param( this->stream,
                                    XINE_PARAM_VO_DEINTERLACE,
                                    ctl->field );
               }
          }
          else {
               config.flags    = DLCONF_OPTIONS;
               config.options &= ~DLOP_DEINTERLACING;
               
               layer->SetConfiguration( layer, &config );
     
               xine_set_param( this->stream,
                               XINE_PARAM_VO_DEINTERLACE,
                               0 );
          }
     }

     if (flags & DCF_SHOW) {
          if (ctl->deinterlace)
               dfx_display_text( this, "Deinterlace: enabled (%s Field)",
                                 (ctl->field == 2) ? "Bottom" : "Top" );
          else
               dfx_display_text( this, "Deinterlace: disabled" );
     }
}


static void*
dfx_control_handler( void *data )
{
     DFXCore              *this     = (DFXCore*) data;
     DFXVideoControl      *ctl      = &this->ctl;
     IDirectFBEventBuffer *event    = ctl->event;
     bool                  acquired = false;
     bool                  monitor  = false;
 
     pthread_setcanceltype( PTHREAD_CANCEL_DEFERRED, NULL );
     
     if (!dfx_control_play( this, 0 )) {
          ctl->active = false;
          return (void*) 1;
     }

     while (1) {
          DFBResult    err;
          DFBUserEvent evt;
          
          pthread_testcancel();
          
          err = event->WaitForEventWithTimeout( event, 0, 100 );
          if (err == DFB_TIMEOUT) {
               if (monitor && dfx_getpl( this )) {
                    ctl->status = xine_get_status( this->stream );
                    ctl->speed  = xine_get_param( this->stream,
                                                  XINE_PARAM_SPEED );
                    dfx_display_status( this, DDF_BOTH );
               }
               continue;
          }
          
          while (event->GetEvent( event, DFB_EVENT(&evt) ) == DFB_OK) {
               int notice = DFX_NOTICE( evt.type );
               int actid  = DFX_ACTION( evt.type );
               int param  = DFX_PARAM( evt.type );
               int delay  = millisec( NULL ) - (unsigned long) evt.data;
          
               pthread_testcancel();

               if (notice == DAN_ACQUIRE) {
                    switch (actid) {
                         case DAI_TOGGLE_PLAY:
                         case DAI_PLAY:
                         case DAI_TOGGLE_PAUSE:
                         case DAI_SEEK:
                         case DAI_SPEED:
                              if (dfx_getpl( this )) {
                                   ctl->status = xine_get_status( this->stream );
                                   ctl->speed  = xine_get_param( this->stream,
                                                            XINE_PARAM_SPEED );
                                   dfx_display_status( this, DDF_BOTH );
                                   monitor = true;
                              }
                              break;

                         case DAI_VOLUME:
                              dfx_control_volume( this, DCF_SHOW, 0 );
                              break;

                         case DAI_BRIGHTNESS:
                         case DAI_CONTRAST:
                         case DAI_SATURATION:
                         case DAI_HUE:
                              dfx_control_correction( this, actid, DCF_SHOW, 0 );
                              break;

                         case DAI_RATIO:
                              dfx_control_ratio( this, DCF_SHOW, 0 );
                              break;
                              
                         case DAI_ZOOM:
                              dfx_control_zoom( this, DCF_SHOW, 0 );
                              break;

                         case DAI_AUDIO_CHANNEL:
                              dfx_control_audio( this, DCF_SHOW, 0 );
                              break;

                         case DAI_SPU_CHANNEL:
                              dfx_control_spu( this, DCF_SHOW, 0 );
                              break;

                         case DAI_DEINTERLACE:
                              dfx_control_deinterlace( this, DCF_SHOW, 0 );
                              break;
                         
                         default:
                              break;
                    }

                    acquired = true;               
               }
               else if (notice == DAN_RELEASE) {
                    switch (actid) {
                         case DAI_TOGGLE_PLAY:
                         case DAI_PLAY:
                         case DAI_TOGGLE_PAUSE:
                         case DAI_SEEK:
                         case DAI_SPEED:
                              monitor = false;
                              break;

                         default:
                              break;
                    }

                    dfx_display_hide( this, 0 );
                    acquired = false;
               } 
               else {
                    switch (actid) {
                         case DAI_QUIT:
                              xine_stop( this->stream );
                              xine_close( this->stream );
                              ctl->active = false;
                              return (void*) 0;

                         case DAI_TOGGLE_PLAY:
                              if (!dfx_control_play( this, 1 )) {
                                   ctl->active = false;
                                   return (void*) 1;
                              }
                              break;

                         case DAI_PLAY:
                              dfx_playlist_skip( this, param );
                              if (!dfx_control_play( this, 0 )) {
                                   ctl->active = false;
                                   return (void*) 1;
                              }
                              break;

                         case DAI_TOGGLE_PAUSE:
                              dfx_control_pause( this, DCF_ALL );
                              break;

                         case DAI_SEEK:
                              if (delay < DELAY_LIMIT)
                                   dfx_control_seek( this, DCF_ALL, param );
                              break;

                         case DAI_SPEED:
                              dfx_control_speed( this, DCF_ALL, param );
                              break;

                         case DAI_VOLUME:
                              if (delay < DELAY_LIMIT)
                                   dfx_control_volume( this, DCF_ALL, param );
                              break;

                         case DAI_BRIGHTNESS:
                         case DAI_CONTRAST:
                         case DAI_SATURATION:
                         case DAI_HUE:
                              if (delay < DELAY_LIMIT)
                                   dfx_control_correction( this, actid,
                                                           DCF_ALL, param );
                              break;

                         case DAI_RATIO:
                              dfx_control_ratio( this, DCF_ALL, param );
                              break;
                              
                         case DAI_ZOOM:
                              dfx_control_zoom( this, DCF_ALL, param );
                              break;

                         case DAI_AUDIO_CHANNEL:
                              dfx_control_audio( this, DCF_ALL, param );
                              break;

                         case DAI_SPU_CHANNEL:
                              dfx_control_spu( this, DCF_ALL, param );
                              break;

                         case DAI_DEINTERLACE:
                              dfx_control_deinterlace( this, DCF_ALL, param );
                              break;

                         case DAI_OSD_HIDE:
                              dfx_display_hide( this, -1 );
                              continue;
                    
                         default:
                              continue;
                    }
                   
                    /* Automatically hide OSD after 'OSD_TIMEOUT' seconds
                     * if key binding doesn't require a modifier.
                     */
                    if (!acquired)
                         dfx_display_hide( this, OSD_TIMEOUT );
               }
          }
     }

     return (void*) 0;
}

static void
dfx_event_listner( void *data, const xine_event_t *event )
{
     DFXCore *this = (DFXCore*) data;

     switch (event->type) {
          case XINE_EVENT_UI_PLAYBACK_FINISHED: {
               DFXMedia *media = &this->media;
               int       magic;

               if (media->cur) {
                    SAY( "playback finished for mrl '%s'\n", media->cur->mrl );
                    
                    media->repeated++;

                    /* end of playlist and no more loops */
                    if (media->cur      == media->last &&
                        media->repeat   >  0           &&
                        media->repeated >= media->repeat)
                         break;
                    
                    magic = DFX_MAGIC( 0, 0, DAI_PLAY, DAP_PLUS );
                    dfx_control_dispatch( this, magic, &event->tv );
               }
          }    break;
          
          default:
               break;
     }
}

void
dfx_control_start( DFXCore *this )
{
     DFXVideoControl *ctl = &this->ctl;
     DFXVideoContext *ctx = &this->ctx;
   
     /* query default color adjustment values */
     ctx->layer->GetColorAdjustment( ctx->layer, &ctl->reset );
     
     if (ctl->deinterlace)
          dfx_control_deinterlace( this, DCF_SET, DAP_FIELD_TOP );
          
     if (ctl->default_ratio)
          dfx_control_ratio( this, DCF_SET, this->ctl.default_ratio );
    
     DFBCHECK(this->dfb->CreateEventBuffer( this->dfb, &ctl->event ));
    
     ctl->active = true;
     if (pthread_create( &ctl->thread, NULL,
                         dfx_control_handler, (void*) this )) 
          FATAL( "couldn't create control thread\n" );
     
     ctl->queue = xine_event_new_queue( this->stream );
     if (ctl->queue)
          xine_event_create_listener_thread( ctl->queue,
                                   dfx_event_listner, (void*) this );
}

void
dfx_control_dispatch( DFXCore *this, int magic, const struct timeval *t )
{
     DFXVideoControl *ctl = &this->ctl;

     if (DFX_EVENT( magic )) {
          xine_event_t evt = {
               .data        = (void*) &evt,
               .data_length = sizeof(xine_event_t),
               .type        = DFX_EVENT( magic ),
               .stream      = this->stream
          };
          
          if (!t)
               gettimeofday( &evt.tv, NULL );
          else
               evt.tv = *t;
          
          xine_event_send( this->stream, &evt );
     } 
     else {
          DFBEvent evt;
          
          evt.clazz     = DFEC_USER;
          evt.user.type = magic;
          evt.user.data = (void*) millisec( t );
          ctl->event->PostEvent( ctl->event, &evt );
     }
}

bool
dfx_control_active( DFXCore *this )
{
     return this->ctl.active;
}

void
dfx_control_release( DFXCore *this )
{
     DFXVideoControl *ctl = &this->ctl;
     DFXVideoContext *ctx = &this->ctx;

     if (ctl->thread) {
          pthread_cancel( ctl->thread );
          pthread_join( ctl->thread, NULL );
          xine_stop( this->stream );
          xine_close( this->stream );
          ctl->active = false;
     }

     if (ctl->queue)
          xine_event_dispose_queue( ctl->queue );

     if (ctl->event)
          ctl->event->Release( ctl->event );

     if (ctl->reset.flags && ctx->layer)
          ctx->layer->SetColorAdjustment( ctx->layer, &ctl->reset );

     dfx_display_release( this );
}

