/*
 * Copyright (C) 2004-2005 Claudio "KLaN" Ciccani <klan@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <math.h>

#include "df_xine.h"


#define CBRT_352  7.060696671 /* cbrt( 352 ) */


#define SEEK_BAR_WIDTH    280 /* width of bar referred to a 352x240 video*/
#define SEEK_BAR_BOTTOM     8 /* bottom margin referred to a 352x240 video */

#define TEXT_TOP            5 /* top margin referred to a 352x240 video */
#define TEXT_LEFT           5 /* left margin referred to a 352x240 video */


#ifdef MIN
# undef MIN
#endif
#define MIN( x, y )       ((x) < (y) ? (x) : (y))

#ifdef MAX
# undef MAX
#endif
#define MAX( x, y )       ((x) > (y) ? (x) : (y))

#ifdef CLAMP
# undef CLAMP
#endif
#define CLAMP( a, b, c )  ((a) < (b)) ? (b) : (((a) > (c)) ? (c) : (a))



static const char *status_syms[] =
{
     [0] = "", /* idle */
     [1] = "}", /* stop */
     [2] = ">", /* play */
     [3] = "{"  /* quit */
};

static const char *speed_syms[]  =
{
     [0] = "<",   /* pause */
     [1] = "@@>", /* slow 4 */
     [2] = "@>",  /* slow 2 */
     [3] = ">",   /* normal */
     [4] = ">$",  /* fast 2 */
     [5] = ">$$"  /* fast 4 */
};



/* XINE_SPEED_* to index */
static inline int
dfx_s2i( int speed )
{
     int i = 0;

     if (!speed)
          return 0;
     
     while (speed >> ++i);

     return i;
}


static bool
dfx_display_acquire( DFXCore *this, int flags )
{
     DFXDisplay *dpy   = &this->dpy;
     int         dst_w = this->ctx.video.width;
     int         dst_h = this->ctx.video.height;

     if (dst_w < 1 || dst_h < 1)
          return false;

     if (dpy->dst_width != dst_w || dpy->dst_height != dst_h) {
          dpy->dst_width  = dst_w;
          dpy->dst_height = dst_h;

          /* this is totally arbitrary */
          switch ((int) cbrt( dst_w * dst_h )) {
               case 0  ... 49:
                    dpy->fontsize = 16;
                    break;
               case 50 ... 69:
                    dpy->fontsize = 20;
                    break;
               default:
                    dpy->fontsize = 24;
                    break;
          }

          DBUG( "using font size %i for OSD\n", dpy->fontsize );
          
          if (dpy->bar.active) {
               xine_osd_hide( dpy->bar.text, 0 );
               xine_osd_hide( dpy->bar.img,  0 );
               dpy->bar.active = 0;
          }

          if (dpy->info.active) {
               xine_osd_hide( dpy->info.text, 0 );
               dpy->info.active = 0;
          }
          
          if (dpy->bar.img) {
               xine_osd_free( dpy->bar.img );
               dpy->bar.img = NULL;
          }

          if (dpy->bar.text) {
               xine_osd_free( dpy->bar.text );
               dpy->bar.text = NULL;
          }

          if (dpy->info.text) {
               xine_osd_free( dpy->info.text );
               dpy->info.text = NULL;
          }
     }
     
     if (flags & DDF_BAR) {
          int x, y;
          
          if (dpy->bar.img && dpy->bar.text)
               return true;
          
          dpy->bar.w = (int) (((double) (dst_w * SEEK_BAR_WIDTH) * CBRT_352)
                               / (cbrt( (double) dst_w ) * 352.0));
          dpy->bar.h = dpy->fontsize + 2;
          x          = (dst_w - dpy->bar.w) >> 1;
          y          = dst_h - dpy->bar.h - (dst_h * SEEK_BAR_BOTTOM / 240);
          
          dpy->bar.img = xine_osd_new( this->stream, 0, 0,
                                       dpy->bar.w + 1, dpy->bar.h + 1 );
          if (!dpy->bar.img)
               return false;

          dpy->bar.text = xine_osd_new( this->stream, 0, 0,
                                        dpy->bar.w + 1, dpy->bar.h + 1 );
          if (!dpy->bar.text) {
               xine_osd_free( dpy->bar.img );
               dpy->bar.img = NULL;
               return false;
          }

          xine_osd_set_font( dpy->bar.text, "sans", dpy->fontsize );
          xine_osd_set_text_palette( dpy->bar.text,
                                     XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT,
                                     XINE_OSD_TEXT1 );

          xine_osd_set_position( dpy->bar.img,  x, y );
          xine_osd_set_position( dpy->bar.text, x, y );
     }

     if (flags & DDF_INFO) {
          int x, y;
          
          if (dpy->info.text)
               return true;

          x = dst_w * TEXT_LEFT / 352;
          y = dst_h * TEXT_TOP  / 240;
          
          dpy->info.text = xine_osd_new( this->stream, 0, 0,
                                         dpy->dst_width - x + 1, 
                                         dpy->fontsize + 8 );
          if (!dpy->info.text)
               return false;

          xine_osd_set_font( dpy->info.text, "cetus", dpy->fontsize );
          xine_osd_set_text_palette( dpy->info.text,
                                     XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT,
                                     XINE_OSD_TEXT1 );

          xine_osd_set_position( dpy->info.text, x, y );
     }
     
     return true;
}

void
dfx_display_bar( DFXCore *this, const char *text, int pos )
{
     DFXDisplay *dpy  = &this->dpy;
     char       *str  = NULL;
     int         len;

     if (!dfx_display_acquire( this, DDF_BAR ))
          return;

     pos = CLAMP( pos, 0, 100 );
     len = dpy->bar.w * pos / 100;

     xine_osd_clear( dpy->bar.img );

     if (len > 0)
          xine_osd_draw_rect( dpy->bar.img, 0, 0, len, dpy->bar.h,
                              XINE_OSD_TEXT1 + 9, 1 );
     if (len < dpy->bar.w)
          xine_osd_draw_rect( dpy->bar.img, len, 0, dpy->bar.w, dpy->bar.h,
                              XINE_OSD_TEXT1 + 7, 1 );

     asprintf( &str, "%s: %3i%%", text, pos );

     if (str) {
          int w, h;
          int x, y;

          xine_osd_clear( dpy->bar.text );
          xine_osd_get_text_size( dpy->bar.text, str, &w, &h );

          if (w > dpy->bar.w) { /* video too small */
               free( str );
               asprintf( &str, "%.3s.: %3i%%", text, pos );
               xine_osd_get_text_size( dpy->bar.text, str, &w, &h );
          }
          
          x = MAX( 0, ((dpy->bar.w - w) >> 1));
          y = MAX( 0, ((dpy->bar.h - h) >> 1));

          xine_osd_draw_text( dpy->bar.text, x, y, str, XINE_OSD_TEXT1 );
          free( str );
     }
     
     xine_osd_show( dpy->bar.img,  0 );
     xine_osd_show( dpy->bar.text, 0 );

     dpy->bar.active = true;
}

void
dfx_display_text( DFXCore *this, const char *text, ... )
{
     DFXDisplay *dpy = &this->dpy;
     char       *str = NULL;
     va_list     ap;

     if (!dfx_display_acquire( this, DDF_INFO ))
          return;

     va_start( ap, text );
     vasprintf( &str, text, ap );
     va_end( ap );

     if (str) {
          xine_osd_clear( dpy->info.text );
          xine_osd_draw_text( dpy->info.text, 0, 0, str, XINE_OSD_TEXT1 );
          xine_osd_show( dpy->info.text, 0 );

          dpy->info.active = true;
          
          free( str );
     }
}

void
dfx_display_status( DFXCore *this, int update )
{
     DFXVideoControl *ctl  = &this->ctl;
     int              secs;

     if (update & DDF_INFO) {
          secs = ctl->pos / 1000;

          if (ctl->status == XINE_STATUS_PLAY)
               dfx_display_text( this, "%-3s %.2i:%.2i:%.2i",
                                 speed_syms[dfx_s2i( ctl->speed )],
                                 secs / 3600, (secs / 60) % 60, secs % 60 );
          else
               dfx_display_text( this, "%s", status_syms[ctl->status] );
     }

     if (update & DDF_BAR) {
          if (ctl->len > 0)
               dfx_display_bar( this, "Position", ctl->pos * 100 / ctl->len );
          else
               dfx_display_bar( this, "Position", 0 );
     }
}

void
dfx_display_hide( DFXCore *this, int udelay )
{
     DFXDisplay *dpy  = &this->dpy;
     int64_t     vpts = 0;

     if (udelay > 0) {
          vpts  = xine_get_current_vpts( this->stream );
          vpts += udelay * 90ll;
     }

     if (dpy->bar.active || udelay < 0) {
          if (dpy->bar.text)
               xine_osd_hide( dpy->bar.text, vpts );
          if (dpy->bar.img)
               xine_osd_hide( dpy->bar.img,  vpts );
     }

     if (dpy->info.active || udelay < 0) {
          if (dpy->info.text)
               xine_osd_hide( dpy->info.text, vpts );
     }

     dpy->bar.active  = false;
     dpy->info.active = false;
}

void
dfx_display_release( DFXCore *this )
{
     DFXDisplay *dpy = &this->dpy;

     if (dpy->bar.img)
          xine_osd_free( dpy->bar.img );

     if (dpy->bar.text)
          xine_osd_free( dpy->bar.text );

     if (dpy->info.text)
          xine_osd_free( dpy->info.text );
}

