/*
 * Copyright (C) 2004-2005 Claudio "KLaN" Ciccani <klan@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include <directfb.h>
#include <directfb_keynames.h>

#include "df_xine.h"
#include "keymap.h"


/*****************************************************************************/

static DFXCore core;

/*****************************************************************************/

static const DirectFBPixelFormatNames(formats);
static const DirectFBKeySymbolNames(keys);

const char*
dfx_format_name( DFBSurfacePixelFormat format )
{
     int index = DFB_PIXELFORMAT_INDEX(format);

     if (index > 0 && index < DFB_NUM_PIXELFORMATS)
          return formats[index].name;
     
     return "NONE";
}

DFBSurfacePixelFormat
dfx_format_id( const char *name )
{
     int i;

     for (i = 0; i < sizeof(formats)/sizeof(formats[0]); i++) {
          if (!strcasecmp( name, formats[i].name ))
               return formats[i].format;
     }

     return DSPF_UNKNOWN;
}

static const char*
dfx_key_name( DFBInputDeviceKeySymbol symbol )
{
     static char buf[2] = { 0, 0 };
     int         i;
     
     switch (symbol) {
          case 'a' ... 'z':
          case 'A' ... 'Z':
               buf[0] = symbol;
               return &buf[0];
          default:
               break;
     }

     for (i = 0; i < sizeof(keys)/sizeof(keys[0]); i++) {
          if (keys[i].symbol == symbol)
               return keys[i].name;
     }
     
     return "NONE";
}

/*****************************************************************************/

void
dfx_exit( int status )
{ 
     dfx_control_release( &core );

     dfx_context_release( &core );

     if (core.input)
          core.input->Release( core.input );
          
     if (core.dfb)
          core.dfb->Release( core.dfb );

     dfx_playlist_free( &core );

     if (core.verbosity)
          fprintf(stderr, "df_xine: exiting...\n" );

     exit( status );
}

static void
dfx_usage( void )
{
     int i;

     puts( "\nUsage:  df_xine [options] MRLs...\n" );
     puts( "    -h  , --help             show this help\n" );
     puts( "    -V  , --version          display version information\n" );
     puts( "    -v N, --verbosity=N      set verbosity level to N" );
     puts( "                             (minimum: 0, maximum: 2, default: 1)\n" );
     puts( "    -l N, --layer=N          use layer N for video playback" );
     puts( "                             (default: auto)\n" );
     puts( "    -p S, --pixelformat=S    set layer pixelformat to S" );
     puts( "                             (default: auto)\n" );
     puts( "    -b S, --buffermode=S     use buffermode S (single, double or triple)" );
     puts( "                             (default: double)\n" );
     puts( "    -d  , --deinterlace      enable frame deinterlacing by default" );
     puts( "                             (default: disabled)\n" );
     puts( "    -f S, --fieldparity=S    set output field parity to S (top or bottom)" );
     puts( "                             (default: none)\n" );
     puts( "    -a N, --aspect=N         override default aspect ratio (e.g. 4:3,1:1,..)" );
     puts( "                             (default: don't override)\n" );
     puts( "    -c S, --crop=S           crop video by top:bottom:left:right" );
     puts( "                             (default: 0:0:0:0)\n" ); 
     puts( "    -s  , --sync             wait for vertical retrace" );
     puts( "                             (default: don't wait)\n" );
     puts( "        , --no-scale         disable scaling the video" );
     puts( "                             (default: enabled)\n" );
     puts( "        , --enable-hwosd     enable hardware On Screen Display when available" );
     puts( "                             (default: disabled)\n" );
     puts( "    -r N, --repeat=N         repeat playlist N times" );
     puts( "                             (default: 1)\n" );
     puts( "    -R  , --recursive        repeat playlist infinite times\n" );
     puts( "        , --stdctl           turn on controlling xine over STDIN\n");
     puts( "    Use:" );

     for (i = 0; i < (sizeof(kmap)/sizeof(kmap[0])); i++) {
          if (kmap[i].event)
               continue;

          if (kmap[i].subkey) {
               printf( "        %-12s + %-12s\tto %s\n",
                       dfx_key_name( kmap[i].key ), 
                       dfx_key_name( kmap[i].subkey ),
                       kmap[i].help );
          }
          else
               printf( "        %-12s   %12c\tto %s\n",
                       dfx_key_name( kmap[i].key ), ' ',  kmap[i].help );
     }

     puts( "\n    For DVD navigation use:" );

     for (i = 0; i < (sizeof(kmap)/sizeof(kmap[0])); i++) {
          if (!kmap[i].event)
               continue;

          if (kmap[i].subkey) {
               printf( "        %-12s + %-12s\tto %s\n",
                       dfx_key_name( kmap[i].key ), 
                       dfx_key_name( kmap[i].subkey ),
                       kmap[i].help );
          }
          else
               printf( "        %-12s   %12c\tto %s\n",
                       dfx_key_name( kmap[i].key ), ' ',  kmap[i].help );
     }

     putchar( '\n' );

     exit( EXIT_FAILURE );
}

static void
dfx_init( DFXCore *this, int *argn, char ***argp )
{
     char **argv;
     int    argc;
     int    i;

     memset( this, 0, sizeof( DFXCore ) );

     this->verbosity      =  1;
     this->stdctl         = false;
     this->scale          = true;
     this->hwosd          = false;
     this->ctx.lid        = -1;
     this->ctx.buffermode = DLBM_BACKVIDEO;
     this->ctx.gmode      = DGM_VIDEO;
     this->media.repeat   =  1;

     DirectFBInit( argn, argp );

     for (argv = *argp, argc = *argn, i = 1; i < argc; i++) {
          char *option = argv[i];

          if (!option || !*option)
               continue;

          if (*option == '-') {
               option++;

               if (!strcmp( option, "h" ) || !strcmp( option, "-help" ))
                    dfx_usage();

               if (*option == 'V' || !strcmp( option, "-version" )) {
                    printf( "df_xine %s is a DirectFB frontend for xine-lib-%s\n",
                            DIRECTFB_VERSION, XINE_VERSION );
                    if (argc == 2)
                         exit( EXIT_SUCCESS );
               }
               else if (*option == 'v') {
                    if (isdigit( option[1] ))
                         option++;
                    else
                    if (argv[i+1] && isdigit( *argv[i+1] ))
                         option = argv[++i];
                    else
                         continue;
                    
                    this->verbosity = *option - '0';
                    DBUG( "verbosity level set to %i\n", this->verbosity );
               }
               else if (!strncmp( option, "-verbosity=", 11 )) {
                    option += 11;
                    if (isdigit( *option )) {
                         this->verbosity = *option - '0';
                         DBUG( "verbosity level set to %i\n", this->verbosity );
                    }
               }
               else if (*option == 'l') {
                    if (isdigit( option[1] ))
                         option++;
                    else
                    if (argv[i+1] && isdigit( *argv[i+1] ))
                         option = argv[++i];
                    else
                         continue;

                    this->ctx.lid = strtol( option, NULL, 10 );
                    DBUG( "will use layer %i\n", this->ctx.lid );
               }
               else if (!strncmp( option, "-layer=", 7 )) {
                    option += 7;
                    if (isdigit( *option )) {
                         this->ctx.lid = strtol( option, NULL, 10 );
                         DBUG( "will use layer %i\n", this->ctx.lid );
                    }
               }
               else if (*option == 'p') {
                    if (isalpha( option[1] ))
                         option++;
                    else
                    if (argv[i+1] && isalpha( *argv[i+1] ))
                         option = argv[++i];
                    else
                         continue;
                    
                    this->ctx.user_format = dfx_format_id( option );
                    DBUG( "suggested pixelformat is %s\n",
                          dfx_format_name( this->ctx.user_format ) );
               }
               else if (!strncmp( option, "-pixelformat=", 13 )) {
                    option += 13;
                    if (*option) {
                         this->ctx.user_format = dfx_format_id( option );
                         DBUG( "suggested pixelformat is %s\n",
                                   dfx_format_name( this->ctx.user_format ) );
                    }
               }                       
               else if (*option == 'b') {
                    if (isalpha( option[1] ))
                         option++;
                    else
                    if (argv[i+1] && isalpha( *argv[i+1] ))
                         option = argv[++i];
                    else
                         continue;
                    
                    if (!strcasecmp( option, "single" )) {
                         this->ctx.buffermode = DLBM_FRONTONLY;
                         DBUG( "buffermode set to %s\n", option );
                    } else
                    if (!strcasecmp( option, "double" )) {
                         this->ctx.buffermode = DLBM_BACKVIDEO;
                         DBUG( "buffermode set to %s\n", option );
                    } else
                    if (!strcasecmp( option, "triple" )) {
                         this->ctx.buffermode = DLBM_TRIPLE;
                         DBUG( "buffermode set to %s\n", option );
                    }
               }
               else if (!strncmp( option, "-buffermode=", 12 )) {
                    option += 12;
                    if (*option) {
                         if (!strcasecmp( option, "single" )) {
                              this->ctx.buffermode = DLBM_FRONTONLY;
                              DBUG( "buffermode set to %s\n", option );
                         } else
                         if (!strcasecmp( option, "double" )) {
                              this->ctx.buffermode = DLBM_BACKVIDEO;
                              DBUG( "buffermode set to %s\n", option );
                         } else
                         if (!strcasecmp( option, "triple" )) {
                              this->ctx.buffermode = DLBM_TRIPLE;
                              DBUG( "buffermode set to %s\n", option );
                         }
                    }
               }
               else if (!strcmp( option, "d" ) || !strcmp( option, "-deinterlace" )) {
                    this->ctl.deinterlace = 1;  
                    DBUG( "deinterlacing enabled by default\n" );
               }
               else if (*option == 'f') {
                    if (isalpha( option[1] ))
                         option++;
                    else
                    if (argv[i+1] && isalpha( *argv[i+1] ))
                         option = argv[++i];
                    else
                         continue;
                         
                    if (!strcasecmp( option, "top" )) {
                         this->ctx.fieldparity = 1;
                         DBUG( "fieldparity set to %s\n", option );
                    } else
                    if (!strcasecmp( option, "bottom" )) {
                         this->ctx.fieldparity = 2;
                         DBUG( "fieldparity set to %s\n", option );
                    }
               }
               else if (!strncmp( option, "-fieldparity=", 13 )) {
                    option += 13;
                    if (*option) {                    
                         if (!strcasecmp( option, "top" )) {
                              this->ctx.fieldparity = 1;
                              DBUG( "fieldparity set to %s\n", option );
                         } else
                         if (!strcasecmp( option, "bottom" )) {
                              this->ctx.fieldparity = 2;
                              DBUG( "fieldparity set to %s\n", option );
                         }
                    }
               }
               else if (*option == 'a') {
                    int num = 0, den = 0;
                    
                    if (isdigit( option[1] ))
                         option++;
                    else if (argv[i+1] && isdigit( *argv[i+1] ))
                         option = argv[++i];
                    else
                         continue;
                    
                    sscanf( option, "%d:%d", &num, &den );
                    if (num && den) {
                         this->ctl.default_ratio = (num << 16) | (den & 0xffff);
                         DBUG( "default aspect ratio set to %d:%d\n", num, den );
                    }
               }
               else if (!strncmp( option, "-aspect=", 8 )) {
                    int num = 0, den = 0;
                    
                    sscanf( option+8, "%d:%d", &num, &den );
                    if (num && den) {
                         this->ctl.default_ratio = (num << 16) | (den & 0xffff);
                         DBUG( "default aspect ratio set to %d:%d\n", num, den );
                    }
               }
               else if (*option == 'c') {
                    DFBInsets c;
                    
                    if (sscanf( option+1, "%d:%d:%d:%d",
                                    &c.t, &c.b, &c.l, &c.r ) == 4) {
                         this->ctx.crop = c;
                    }
                    else if (sscanf( argv[i+1], "%d:%d:%d:%d",
                                    &c.t, &c.b, &c.l, &c.r ) == 4) {
                         this->ctx.crop = c;
                         i++;
                    }
                    else
                         continue;

                    DBUG( "crop set to %d:%d:%d:%d\n", c.t, c.b, c.l, c.r );
               }
               else if (!strncmp( option, "-crop=", 6 )) {
                    DFBInsets c;
                    
                    if (sscanf( option+6, "%d:%d:%d:%d",
                                    &c.t, &c.b, &c.l, &c.r ) == 4) {
                         this->ctx.crop = c;
                         DBUG( "crop set to %d:%d:%d:%d\n",
                               c.t, c.b, c.l, c.r );
                    }
               }
               else if (!strcmp( option, "s" ) || !strcmp( option, "-sync" )) {
                    this->ctx.flipflags = DSFLIP_WAITFORSYNC;
                    DBUG( "will wait for vertical retrace after flipping\n" );
               }
               else if (!strcmp( option, "-no-scale" )) {
                    this->scale = false;
                    DBUG( "vide scaling disabled\n" );
               }
               else if (!strcmp( option, "-enable-hwosd" )) {
                    this->hwosd = true;
                    DBUG( "hardware OSD enabled\n" );
               }
               else if (!strcmp( option, "R" ) || !strcmp( option, "-recursive" )) {
                    this->media.repeat = -1;
                    DBUG( "playlist loop set to recursive\n" );
               }
               else if (*option == 'r') {
                    if (isdigit( option[1] ))
                         option++;
                    else
                    if (argv[i+1] && isdigit( *argv[i+1] ))
                         option = argv[++i];
                    else
                         continue;
                    
                    this->media.repeat = strtol( option, NULL, 10 );
                    DBUG( "playlist will be repeated %i times\n",
                          this->media.repeat );
               }
               else if (!strncmp( option, "-repeat=", 8 )) {
                    option += 8;
                    if (isdigit( *option )) {
                         this->media.repeat = strtol( option, NULL, 10 );
                         DBUG( "playlist will be repeated %i times\n",
                                this->media.repeat );
                    }
               }
	          else if (!strncmp( option, "-stdctl" ,7)) {
		           this->stdctl = 1;
                     DBUG( "standard input control enabled\n" );
               }
          }
          else
               dfx_playlist_append( this, option );
     }

     /* is playlist empty ? */
     if (!this->media.first)
          FATAL( "no MRL selected, playlist is empty\n" );
}


static inline int
kmap_lookup_key( int key )
{
     int i;

     for (i = 0; i < (sizeof(kmap)/sizeof(kmap[0])); i++) {
          if (kmap[i].key == key)
               return i;
     }

     return -1;
}

static inline int
kmap_lookup_keycomb( int key, int subkey )
{
     int i;

     for (i = 0; i < (sizeof(kmap)/sizeof(kmap[0])); i++) {
          if (kmap[i].key == key && kmap[i].subkey == subkey)
               return i;
     }

     return -1;
}

static int
dfx_handle_keypress( DFXCore *this, int key, DFBInputEvent *evt )
{
     int id;
     int subkey;
     int magic;

     if (key == -1) {
          key = tolower( evt->key_symbol );
          id  = kmap_lookup_key( key );

          if (id < 0)
               return -1;

          if (!kmap[id].subkey) {
               magic = DFX_MAGIC( kmap[id].event, 0,
                                  kmap[id].actid, kmap[id].param );
               key   = -1;
          }
          else 
               magic = DFX_MAGIC( kmap[id].event, DAN_ACQUIRE,
                                  kmap[id].actid, kmap[id].param );
                                  
          dfx_control_dispatch( this, magic, &evt->timestamp );
     } else {
          subkey = tolower( evt->key_symbol );
          id     = kmap_lookup_keycomb( key, subkey );

          if (id < 0)
               return key;

          magic = DFX_MAGIC( kmap[id].event, 0,
                             kmap[id].actid, kmap[id].param );
          dfx_control_dispatch( this, magic, &evt->timestamp );
     }

     return key;
}

static int
dfx_handle_keyrelease( DFXCore *this, int key, DFBInputEvent *evt )
{
     int id, magic;

     if (key == tolower( evt->key_symbol )) {
          id = kmap_lookup_key( key );
          
          magic = DFX_MAGIC( kmap[id].event, DAN_RELEASE,
                             kmap[id].actid, kmap[id].param );
          dfx_control_dispatch( this, magic, &evt->timestamp );

          key = -1;
     }

     return key;
}

static void
dfx_event_loop( DFXCore *this )
{
     IDirectFBEventBuffer *input;
     
     DFBCHECK(this->dfb->CreateInputEventBuffer( this->dfb,
                                                 DICAPS_KEYS, 1, &input ));
     this->input = input;

     while (dfx_control_active( this )) {
          DFBInputEvent evt;
          static int    key = -1;
          
          input->WaitForEventWithTimeout( input, 0, 10 );

          while (input->GetEvent( input, DFB_EVENT( &evt) ) == DFB_OK) {
               if (!dfx_control_active( this ))
                    return;
               
               switch (evt.type) {
                    case DIET_KEYPRESS:
                         key = dfx_handle_keypress( this, key, &evt );
                         break;
                    case DIET_KEYRELEASE:
                         key = dfx_handle_keyrelease( this, key, &evt );
                         break;
                    default:
                         break;
               }
          }
     }
}


/*
 * Chris Griffiths <chris@y-fronts.com>:
 *
 *   Alternate main loop - reads commands from stdin, uses keymap.h to determine
 *   required action and requests it. Used to run xine from within 
 *   freevo (www.freevo.org).
 */

static inline int
dfx_stdctl_get_action_magic( char action_name[255] )
{
     int id;

     for (id = 0; id < (sizeof(kmap)/sizeof(kmap[0])); id++) {
          if (kmap[id].stdctl && !strcasecmp( kmap[id].stdctl, action_name ))
               return DFX_MAGIC( kmap[id].event, 0, kmap[id].actid, kmap[id].param );
     }
     
     return -1;
}

static void
dfx_stdctl_loop( DFXCore *this ) 
{
     char            c[255];
     int             len;
     int             key_magic;
     fd_set          readset;
     struct timeval  timeout;

     while (dfx_control_active( this )) {
          timeout.tv_sec  =  0;
          timeout.tv_usec =  500000;

          FD_ZERO( &readset );
          FD_SET( STDIN_FILENO, &readset );
          select( STDIN_FILENO + 1, &readset, NULL, NULL, &timeout );

          if (FD_ISSET( STDIN_FILENO, &readset )) {
               len = read( STDIN_FILENO, &c, 255 );
               if (len > 0) {
                    c[len-1] = '\0';
                    key_magic = dfx_stdctl_get_action_magic( c );
                    if (key_magic != -1) {
                         struct timeval tv;
                         gettimeofday( &tv, NULL );
                         dfx_control_dispatch( this, key_magic, &tv );
                    }
               }
          }
     }
}


int
main( int argc, char *argv[] )
{
     DFXCore *this = &core;
     
     dfx_init( this, &argc, &argv);

     DFBCHECK(DirectFBCreate( &this->dfb )); 
     SAY( "here we go!\n" );
     
     this->dfb->GetDeviceDescription( this->dfb, &this->card_caps );
     if (!this->hwosd)
          this->card_caps.blitting_flags &= ~DSBLIT_BLEND_ALPHACHANNEL;     
 
     dfx_context_create( this );
     dfx_control_start( this );

     if (this->stdctl) 
          dfx_stdctl_loop( this );
     else
          dfx_event_loop( this );
     
     dfx_exit( EXIT_SUCCESS );
     
     return 0;
}

