/*
 * Copyright (C) 2004-2005 Claudio "KLaN" Ciccani <klan@users.sf.net>
 *
 * This file is subject to the terms and conditions of the MIT License:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <pthread.h>

#include <directfb.h>
#include <directfb_strings.h>

#include <direct/types.h>
#include <direct/messages.h>
#include <direct/mem.h>
#include <direct/hash.h>

#include "dfblogo.h"


#define DFBCHECK( x... ) \
{\
     int err = x;\
     if (err != DFB_OK) {\
          fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ );\
          DirectFBErrorFatal( #x, err );\
     }\
}

#ifdef min
# undef min
#endif
#define min( a, b )  (((a)<(b)) ? (a) : (b))

#ifdef max
# undef max
#endif
#define max( a, b )  (((a)>(b)) ? (a) : (b))

#ifdef CLAMP
# undef CLAMP
#endif
#define CLAMP( a, min, max )  (((a)<(min)) ? (min) : (((a)>(max)) ? (max) : (a)))


/****************************** Declared Types ********************************/

struct stack_entry {
     IDirectFBWindow   *window;
     IDirectFBSurface  *surface;
     bool               stopped;
};

struct frame_data {
     IDirectFBSurface  *frame;
     IDirectFBSurface  *logo;
     DirectHash        *stack;
     pthread_mutex_t    mutex;
     int                progress;
};

struct blitter_data {
     struct frame_data *frame_data;
     int                sw;
     int                sh;
     DFBColor          *color;
     DFBRectangle       rect[2];
};

/****************************** Global Options ********************************/

static DFBSurfacePixelFormat  format    = DSPF_UNKNOWN;
static int                    n_windows = 1;
static int                    use_logo  = 1;
static char                  *mrl       = NULL;

/******************************************************************************/


static void
usage( void )
{
     puts( "Usage: sample1 [options] video\n" );
     puts( "       -h,--help               show this help" );
     puts( "       --format=<pixelformat>  select the pixelformat to use" );
     puts( "       --windows=N             play video on N windows (default:1, maximum:20)" );
     puts( "       --no-logo               do not display DirectFB logo in the" );
     puts( "                               lower-left corner of the window\n" );
     puts( "       Use:" );
     puts( "           ESC,Q,q   to quit" );
     puts( "           Enter     to toggle pause globally" );
     puts( "           Space     to toggle pause on the captured window" );
     puts( "           ->,<-     to seek" );
     puts( "           B,b + ->  to increase brightness" );
     puts( "           B,b + <-  to decrease brightness" );
     puts( "           C,c + ->  to increase contrast" );
     puts( "           C,c + <-  to decrease contrast" );
     puts( "           S,s + ->  to increase saturation" );
     puts( "           S,s + <-  to decrease saturation" );
     puts( "           H,h + ->  to increase hue" );
     puts( "           H,h + <-  to decrease hue\n" );
     exit( EXIT_FAILURE );
}

static void
parse_options( int argc, char **argv )
{
     int i;
     
     for (i = 1; i < argc; i++) {
          char *option = argv[i];

          if (!option || !*option)
               continue;

          if (*option == '-') {
               option++;

               if (!strcmp( option, "h" ) || !strcmp( option, "-help" ))
                    usage();

               if (!strncmp( option, "-format=", sizeof("-format=")-1 )) {
                    DirectFBPixelFormatNames(formats);
                    int j;

                    option += sizeof("-format=")-1;
                    if (!isalpha( *option ))
                         continue;

                    for (j = 0; j < sizeof(formats)/sizeof(formats[0]); j++) {
                         if (!strcasecmp( option, formats[j].name )) {
                              format = formats[j].format;
                              D_INFO( "sample1: using pixelformat %s.\n",
                                       formats[j].name );
                              break;
                         }
                    }

                    if (!format) {
                         D_ERROR( "sample1: unknown pixelformat \"%s\"!\n", option );
                         exit( EXIT_FAILURE );
                    }
               } else
               if (!strncmp( option, "-windows=", sizeof("-windows=")-1 )) {
                    option += sizeof("-windows=")-1;
                    if (isdigit( *option )) {
                         n_windows = strtol( option, NULL, 10 );
                         n_windows = CLAMP( n_windows, 1, 20 );
                    }
                    D_INFO( "sample1: using %i windows.\n", n_windows );
               }
               if (!strcmp( option, "-no-logo" )) {
                    use_logo = 0;
                    D_INFO( "sample1: logo disabled.\n" );
               }
               
          } else 
          if (i == argc-1) {
               mrl = option;
          }
     }

     if (!mrl || !*mrl)
          usage();
}


static void
dump_stream_info( IDirectFBVideoProvider *video, const char *filename )
{
     DFBStreamDescription desc;

     if (video->GetStreamDescription( video, &desc ))
          return;

     printf( "\n%s:\n", filename );
     
     if (desc.caps & DVSCAPS_VIDEO) {
          DFBSurfaceDescription dsc;

          video->GetSurfaceDescription( video, &dsc );
          
          printf( "  # Video: %s, %dx%d (ratio %.3f), %.2ffps, %dKbits/s\n",
                  *desc.video.encoding ? desc.video.encoding : "Unknown",
                  dsc.width, dsc.height, desc.video.aspect, 
                  desc.video.framerate, desc.video.bitrate / 1000 );
     }

     if (desc.caps & DVSCAPS_AUDIO) {
          printf( "  # Audio: %s, %dKhz, %d channel(s), %dKbits/s\n",
                  *desc.audio.encoding ? desc.audio.encoding : "Unknown",
                  desc.audio.samplerate / 1000, desc.audio.channels,
                  desc.audio.bitrate / 1000 );
     }

     printf( " Title:   %s\n"
             " Author:  %s\n"
             " Album:   %s\n"
             " Year:    %d\n"
             " Genre:   %s\n"
             " Comment: %s\n\n",
             desc.title, desc.author, desc.album,
             desc.year, desc.genre, desc.comment );

     fflush( stdout );
}


static void
create_logo( IDirectFB         *dfb,
             IDirectFBSurface **ret_logo )
{
     IDirectFBSurface             *logo = NULL;
     DFBGraphicsDeviceDescription  caps;
     DFBResult                     err;

     dfb->GetDeviceDescription( dfb, &caps );
     DFBCHECK( dfb->CreateSurface( dfb, &dfblogo_desc, &logo ) );

     /* use hardware modulation when available */
     if ((caps.blitting_flags & DSBLIT_COLORIZE) &&
         (caps.blitting_flags & DSBLIT_BLEND_ALPHACHANNEL)) {
          IDirectFBSurface      *tmp = NULL;
          DFBSurfaceDescription  dsc;

          dsc.flags       = DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_PIXELFORMAT;
          dsc.width       = dfblogo_desc.width;
          dsc.height      = dfblogo_desc.height;
          dsc.pixelformat = dfblogo_desc.pixelformat;
               
          err = dfb->CreateSurface( dfb, &dsc, &tmp );
          if (err == DFB_OK && tmp) {
               tmp->Blit( tmp, logo, NULL, 0, 0 );
               logo->Release( logo );
               logo = tmp;
          } else
               DirectFBError( "couldn't create logo surface", err );
     }

     *ret_logo = logo;
}


static void
create_stack( IDirectFBDisplayLayer  *layer,
              IDirectFBEventBuffer   *buffer,
              DFBSurfaceDescription   video_dsc,
              DirectHash            **ret_stack )
{
     DirectHash           *stack;
     DFBWindowDescription  dsc;
     int                   i;
     
     /* create hash tables */
     DFBCHECK( direct_hash_create( n_windows, &stack ) );

     dsc.flags  = DWDESC_POSX | DWDESC_POSY | DWDESC_WIDTH | DWDESC_HEIGHT;
     dsc.posx   = 0;
     dsc.posy   = 0;
     dsc.width  = video_dsc.width;
     dsc.height = video_dsc.height;
     if (format) {
          dsc.flags       |= DWDESC_PIXELFORMAT;
          dsc.pixelformat  = format;
     }

     for (i = 0; i < n_windows; i++, dsc.posx += 10, dsc.posy += 5) {
          struct stack_entry *entry;
          IDirectFBWindow    *window; 
          DFBWindowID         id;
          IDirectFBSurface   *surface;
          
          entry = D_CALLOC( 1, sizeof(struct stack_entry) );
          if (!entry) { 
               D_ERROR( "sample1: out of memory!!\n" );
               exit( EXIT_FAILURE );
          }
               
          DFBCHECK( layer->CreateWindow( layer, &dsc, &window ) );
          window->GetID( window, &id );
          window->SetOpacity( window, 0xff );
          window->AttachEventBuffer( window, buffer );
          window->GetSurface( window, &surface );
          
          surface->Clear( surface, 0x00, 0x00, 0x00, 0xff );
          surface->Flip( surface, NULL, 0 );
          
          entry->window  = window;
          entry->surface = surface;
          DFBCHECK( direct_hash_insert( stack, id, (void*) entry ) );
     }

     *ret_stack = stack;
}

static IDirectFBWindow*
lookup_window( DirectHash   *stack,
               DFBWindowID   id )
{
     struct stack_entry* entry;

     entry = (struct stack_entry*) direct_hash_lookup( stack, id );
     if (entry)
          return entry->window;

     return NULL;
}

static bool
toggle_window( DirectHash   *stack,
               DFBWindowID   id )
{
     struct stack_entry* entry;

     entry = (struct stack_entry*) direct_hash_lookup( stack, id );
     if (entry) {
          entry->stopped ^= 1;
          return entry->stopped;
     }

     return false;
}

static IDirectFBWindow*
remove_window( DirectHash   *stack,
               DFBWindowID   id )
{
     struct stack_entry *entry;
     IDirectFBWindow    *window = NULL;

     entry = (struct stack_entry*) direct_hash_lookup( stack, id );
     if (entry) {
          window = entry->window;
          entry->surface->Release( entry->surface );
          window->Destroy( window );
          window->Release( window );
          D_FREE( entry );
          direct_hash_remove( stack, id );
     }

     return window;
}

static bool
stack_destructor( DirectHash *stack,
                  __u32       id,
                  void       *value,
                  void       *ctx )
{
     struct stack_entry *entry = (struct stack_entry*) value;

     entry->surface->Release( entry->surface );
     entry->window->Release( entry->window );
     D_FREE( entry );

     return true;
}

static void
destroy_stack( DirectHash *stack )
{
     direct_hash_iterate( stack, stack_destructor, NULL );
     direct_hash_destroy( stack );
}


static void
create_frame( IDirectFB              *dfb,
              DFBSurfaceDescription   dsc,
              IDirectFBSurface      **ret_frame )
{
     IDirectFBSurface *frame = NULL;
     DFBResult         err;
     
     dsc.flags = DSDESC_WIDTH | DSDESC_HEIGHT;

     if (format) {
          dsc.flags       |= DSDESC_PIXELFORMAT;
          dsc.pixelformat  = format;
     }

     err = dfb->CreateSurface( dfb, &dsc, &frame );
     if (err != DFB_OK)
          DirectFBErrorFatal( "failed to create buffer for video frames", err );

     *ret_frame = frame;
}


static bool
frame_blitter( DirectHash *hash,
               __u32       id,
               void       *value,
               void       *ctx )
{
     struct blitter_data *blt   = (struct blitter_data*) ctx;
     struct stack_entry  *entry = (struct stack_entry*) value;
     IDirectFBSurface    *frame = blt->frame_data->frame;
     IDirectFBSurface    *logo  = blt->frame_data->logo;
     IDirectFBSurface    *dst   = entry->surface;
     DFBColor            *color = blt->color;
     int                  dw,dh;

     if (entry->stopped)
          return true;
     
     dst->GetSize( dst, &dw, &dh );
     dst->SetBlittingFlags( dst, DSBLIT_NOFX );
          
     if (dw != blt->sw || dh != blt->sh)
          dst->StretchBlit( dst, frame, NULL, NULL );
     else
          dst->Blit( dst, frame, NULL, 0, 0 );

     /* draw progressive logo */
     if (logo) {
          /* elapsed */
          dst->SetColor( dst, color->r, color->g, color->b, 0xff );
          dst->SetBlittingFlags( dst, DSBLIT_COLORIZE |
                                      DSBLIT_BLEND_ALPHACHANNEL );
          dst->Blit( dst, logo, &blt->rect[0],
                     7, dh - dfblogo_desc.height - 7 );
          /* remaining */
          dst->SetBlittingFlags( dst, DSBLIT_BLEND_ALPHACHANNEL );
          dst->Blit( dst, logo, &blt->rect[1],
                     7 + blt->rect[0].w, dh - dfblogo_desc.height - 7 );
     }

     dst->Flip( dst, NULL, 0 );
     
     return true;
}

static int
frame_cb( void *cdata )
{
	static DFBColor      color = { 0x22, 0x33, 0xbb, 0xff };
     struct frame_data   *data  = (struct frame_data*) cdata;
     struct blitter_data  blt;
          
     blt.frame_data = data;
     data->frame->GetSize( data->frame, &blt.sw, &blt.sh );

     /* setup coordinates for progressive logo */
     if (data->logo) {
          blt.color     = &color;
          blt.rect[0].y = blt.rect[1].y = 0;
          blt.rect[0].h = blt.rect[1].h = dfblogo_desc.height;
          /* elapsed part */
          blt.rect[0].x = 0;
          blt.rect[0].w = dfblogo_desc.width * data->progress / 100;
          /* remainig part */
          blt.rect[1].x = blt.rect[0].w;
          blt.rect[1].w = dfblogo_desc.width - blt.rect[0].w;
     }
     
     /* recursively blit frame to windows */
     pthread_mutex_lock( &data->mutex );
     direct_hash_iterate( data->stack, frame_blitter, (void*) &blt );
     pthread_mutex_unlock( &data->mutex );

     /* rotate colors */
     if (data->logo) {
          color.r -= 2;
          color.g += 1;
          color.b -= 2;
     }

     return 0;
}


static void
seek( IDirectFBVideoProvider *video,
      double                  step )
{
     double len = 0.0, pos = 0.0;

     video->GetLength( video, &len );
     video->GetPos( video, &pos );
     
     pos += step;
     if (pos < 0)
          pos = 0;
     else if (pos > len)
          return;

     video->SeekTo( video, pos );
}

static void
adjust_color( IDirectFBVideoProvider  *video,
              DFBColorAdjustmentFlags  flags,
              int                      step )
{
     DFBColorAdjustment adj = { .flags = 0 };

     if (video->GetColorAdjustment( video, &adj ) != DFB_OK)
          return;

     if (flags & DCAF_BRIGHTNESS) {
          adj.flags      |= DCAF_BRIGHTNESS;
          adj.brightness  = CLAMP( adj.brightness+step, 0, 0xffff );
     }

     if (flags & DCAF_CONTRAST) {
          adj.flags      |= DCAF_CONTRAST;
          adj.contrast    = CLAMP( adj.contrast+step, 0, 0xffff );
     }

     if (flags & DCAF_HUE) {
          adj.flags      |= DCAF_HUE;
          adj.hue         = CLAMP( adj.hue+step, 0, 0xffff );
     }

     if (flags & DCAF_SATURATION) {
          adj.flags      |= DCAF_SATURATION;
          adj.saturation  = CLAMP( adj.saturation+step, 0, 0xffff );
     }

     if (adj.flags)
          video->SetColorAdjustment( video, &adj );
}


int
main( int argc, char** argv )
{
     IDirectFB                    *dfb          = NULL; 
     IDirectFBDisplayLayer        *layer        = NULL;
     IDirectFBEventBuffer         *buffer       = NULL;
     IDirectFBVideoProvider       *video        = NULL;
     IDirectFBSurface             *frame        = NULL;
     IDirectFBSurface             *logo         = NULL;
     DirectHash                   *window_stack = NULL;
     DFBVideoProviderCapabilities  caps;
     DFBSurfaceDescription         dsc;
     static struct frame_data      data;

     if (argc < 2)
          usage();

     DFBCHECK( DirectFBInit( &argc, &argv ) );
     parse_options( argc, argv );

     DFBCHECK( DirectFBCreate( &dfb ) );
     DFBCHECK( dfb->CreateVideoProvider( dfb, mrl, &video ) );
     video->GetCapabilities( video, &caps );
     video->GetSurfaceDescription( video, &dsc );
  
     /* dump stream informations to stdout */
     dump_stream_info( video, mrl );
     
     if (use_logo)
          create_logo( dfb, &logo );

     DFBCHECK( dfb->GetDisplayLayer( dfb, DLID_PRIMARY, &layer ) );
     DFBCHECK( dfb->CreateEventBuffer( dfb, &buffer ) );
     
     /* create window stack */
     create_stack( layer, buffer, dsc, &window_stack );

     /* create buffer for video frames */
     create_frame( dfb, dsc, &frame );

     data.frame    = frame;
     data.logo     = logo;
     data.stack    = window_stack;
     data.progress = 0;
     pthread_mutex_init( &data.mutex, NULL );
     
     DFBCHECK( video->PlayTo( video, frame, NULL, frame_cb, (void*) &data ) );

     while (1) {
          DFBWindowEvent         evt;
          DFBVideoProviderStatus status = DVSTATE_UNKNOWN;
          static int             paused = 0;
          static int             adj    = 0;

          if (buffer->WaitForEventWithTimeout( buffer, 0, 100 ) == DFB_TIMEOUT) {
               double len = 0.0;
               double pos = 0.0;
               int    w,h;
               
               /* check for video resize event */
               video->GetSurfaceDescription( video, &dsc );
               frame->GetSize( frame, &w, &h );
               if (dsc.width != w || dsc.height != h) {
                    frame->Release( frame ); 
                    create_frame( dfb, dsc, &frame );
                    data.frame = frame;
               }

               /* check for end of playback */
               video->GetStatus( video, &status );
               
               if (status == DVSTATE_FINISHED ||
                   dsc.width != w || dsc.height != h) 
               {
                    DFBCHECK(video->PlayTo( video, frame, NULL,
                                            frame_cb, (void*) &data ));
               }

               video->GetLength( video, &len );
               video->GetPos( video, &pos );
               if (len > 0.0 && pos > 0.0)
                    data.progress = (pos+0.5) * 100.0 / len;
               else
                    data.progress = 0;
               
               continue;
          }

          while (buffer->GetEvent( buffer, DFB_EVENT(&evt) ) == DFB_OK) {
               IDirectFBWindow *window;
               
               switch (evt.type) {
                    case DWET_KEYDOWN:
                         switch (tolower(evt.key_symbol)) {
                              case DIKS_ESCAPE:
                              case DIKS_SMALL_Q:
                                   goto end;

                              case DIKS_ENTER:
                                   video->GetStatus( video, &status );
                                   if (status != DVSTATE_PLAY)
                                        video->PlayTo( video, frame, NULL,
                                                       frame_cb, (void*) &data );
                                   else
                                        video->Stop( video );
                                   break;

                              case DIKS_SPACE:
                                   if (toggle_window( window_stack,
                                                      evt.window_id )) {
                                        /* paused */
                                        paused++;
                                        if (paused >= n_windows) {
                                             video->Stop( video ); 
                                             paused = n_windows;
                                        }
                                   } else {
                                        /* playing */
                                        video->GetStatus( video, &status );
                                        if (status != DVSTATE_PLAY) {
                                             video->PlayTo( video, frame, NULL,
                                                            frame_cb, (void*) &data );
                                        }
                                        paused--;
                                   }
                                   break;                                        

                              case DIKS_SMALL_B:
                                   adj |= DCAF_BRIGHTNESS;
                                   break;
                              case DIKS_SMALL_C:
                                   adj |= DCAF_CONTRAST;
                                   break;
                              case DIKS_SMALL_H:
                                   adj |= DCAF_HUE;
                                   break;
                              case DIKS_SMALL_S:
                                   adj |= DCAF_SATURATION;
                                   break;

                              case DIKS_CURSOR_LEFT:
                                   if (adj)
                                        adjust_color( video, adj, -257 );
                                   else
                                        seek( video, -10.0 );
                                   break;

                              case DIKS_CURSOR_RIGHT:
                                   if (adj)
                                        adjust_color( video, adj, +257 );
                                   else
                                        seek( video, +10.0 );
                                   break;
                              
                              default:
                                   break;
                         }
                         break;

                    case DWET_KEYUP:
                         switch (tolower(evt.key_symbol)) {
                              case DIKS_SMALL_B:
                                   adj &= ~DCAF_BRIGHTNESS;
                                   break;
                              case DIKS_SMALL_C:
                                   adj &= ~DCAF_CONTRAST;
                                   break;
                              case DIKS_SMALL_H:
                                   adj &= ~DCAF_HUE;
                                   break;
                              case DIKS_SMALL_S:
                                   adj &= ~DCAF_SATURATION;
                                   break;
                              default:
                                   break;
                         }
                         break;

                    case DWET_ENTER:
                    case DWET_LEAVE:
                    case DWET_MOTION:
                    case DWET_BUTTONDOWN:
                    case DWET_BUTTONUP:
                         /* limit events to send to mouse events */
                         if (caps & DVCAPS_INTERACTIVE) {
                              int w, h;
                              
                              window = lookup_window( window_stack, evt.window_id );
                              window->GetSize( window, &w, &h );          
                              /* scale window coordinates to video coordinates */
                              evt.x = evt.x * dsc.width  / w;
                              evt.y = evt.y * dsc.height / h;
                         
                              video->SendEvent( video, DFB_EVENT(&evt) );
                         }
                         break;

                    case DWET_CLOSE:
                    case DWET_DESTROYED:
                         pthread_mutex_lock( &data.mutex );
                         window = remove_window( window_stack, evt.window_id );
                         if (window) {
                              if (--n_windows <= 0) {
                                   pthread_mutex_unlock( &data.mutex );
                                   goto end;
                              }
                         }
                         pthread_mutex_unlock( &data.mutex );
                         break;

                    default:
                         break;
               }
          }
     }

end:
     if (video)
          video->Release( video );

     if (buffer)
          buffer->Release( buffer );

     if (window_stack)
          destroy_stack( window_stack );

     if (logo)
          logo->Release( logo );

     if (frame)
          frame->Release( frame );

     if (layer)
          layer->Release( layer );

     if (dfb)
          dfb->Release( dfb );

     pthread_mutex_destroy( &data.mutex );

     return 0;
}

