/*
 * Copyright (C) 2004-2005 Claudio "KLaN" Ciccani <klan@users.sf.net>
 *
 * This file is subject to the terms and conditions of the MIT License:
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>

#include <pthread.h>

#include <directfb.h>
#include <directfb_strings.h>

#include <direct/types.h>
#include <direct/messages.h>
#include <direct/mem.h>

#include <xine.h>
#include <xine/xineutils.h>

#include "dfblogo.h"


#define DFBCHECK( x... ) \
{\
     int err = x;\
     if (err != DFB_OK) {\
          fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ );\
          DirectFBErrorFatal( #x, err );\
     }\
}

#ifdef min
# undef min
#endif
#define min( a, b )  (((a)<(b)) ? (a) : (b))

#ifdef max
# undef max
#endif
#define max( a, b )  (((a)>(b)) ? (a) : (b))

#ifdef CLAMP
# undef CLAMP
#endif
#define CLAMP( a, min, max )  (((a)<(min)) ? (min) : (((a)>(max)) ? (max) : (a)))



/************************** Driver Specific Data ******************************/

typedef void (*DVOutputCallback) ( void                  *cdata,
                                   int                    width,
                                   int                    height,
                                   double                 ratio,
                                   DFBSurfacePixelFormat  format,
                                   DFBRectangle          *dest_rect );

typedef struct {
     IDirectFBSurface *destination;
     IDirectFBSurface *subpicture;

     DVOutputCallback  output_cb;
     void             *output_cdata;

     DVFrameCallback  frame_cb;
     void            *frame_cdata;
} dfb_visual_t;

/****************************** Declared Types ********************************/

struct frame_data {
     IDirectFBWindow   *window;
     IDirectFBSurface  *frame;
     IDirectFBSurface  *osd;
     IDirectFBSurface  *logo;
     IDirectFBSurface  *dest;
     int                progress;
     double             screen_ratio;
     int                width;
     int                height;
     pthread_mutex_t    mutex;
};

/****************************** Global Options ********************************/

static DFBSurfacePixelFormat  format       = DSPF_UNKNOWN;
static bool                   use_hwosd    = true;
static int                    osd_mode     = 0; // 0: texture, 1: window
static bool                   use_logo     = true;
static char                  *mrl          = NULL;

/******************************************************************************/



static void
usage( void )
{
     puts( "Usage: sample2 [options] file\n" );
     puts( "       -h,--help               show this help" );
     puts( "       --format=<pixelformat>  select the pixelformat to use" );
     puts( "       --no-hwosd              disable hardware On Screen Display" );
     puts( "       --osd-mode=<mode>       select hardware On Screen Display mode:" );
     puts( "                                  - texture: a surface will be blended" );
     puts( "                                             together with each video frame" );
     puts( "                                  - window:  a transparent window will be used" );
     puts( "       --no-logo               do not display DirectFB logo in the" );
     puts( "                               lower-left corner of the window\n" );
     puts( "       Use:" );
     puts( "           ESC,Q,q   to quit" );
     puts( "           Enter     to toggle pause" );
     puts( "           ->,<-     to seek" );
     puts( "           B,b + ->  to increase brightness" );
     puts( "           B,b + <-  to decrease brightness" );
     puts( "           C,c + ->  to increase contrast" );
     puts( "           C,c + <-  to decrease contrast" );
     puts( "           S,s + ->  to increase saturation" );
     puts( "           S,s + <-  to decrease saturation" );
     exit( EXIT_FAILURE );
}

static void
parse_options( int argc, char **argv )
{
     int i;
     
     for (i = 1; i < argc; i++) {
          char *option = argv[i];

          if (!option || !*option)
               continue;

          if (*option == '-') {
               option++;

               if (!strcmp( option, "h" ) || !strcmp( option, "-help" ))
                    usage();

               if (!strncmp( option, "-format=", sizeof("-format=")-1 )) {
                    DirectFBPixelFormatNames(formats);
                    int j;

                    option += sizeof("-format=")-1;
                    if (!isalpha( *option ))
                         continue;

                    for (j = 0; j < sizeof(formats)/sizeof(formats[0]); j++) {
                         if (!strcasecmp( option, formats[j].name )) {
                              format = formats[j].format;
                              D_INFO( "sample2: using pixelformat %s.\n",
                                       formats[j].name );
                              break;
                         }
                    }

                    if (!format) {
                         D_ERROR( "sample2: unknown pixelformat '%s'.\n", option );
                         exit( EXIT_FAILURE );
                    }
               } else
               if (!strcmp( option, "-no-hwosd" )) {
                    use_hwosd = false;
                    osd_mode  = 0;
                    D_INFO( "sample2: hardware OSD disabled.\n" );
               } else 
               if (!strncmp( option, "-osd-mode=", sizeof("-osd-mode=")-1) ) {
                    option += sizeof("-osd-mode=")-1;
                    
                    if (isalpha( *option ) && use_hwosd) {
                         if      (!strcmp( option, "texture" ))
                              osd_mode = 0;
                         else if (!strcmp( option, "window" ))
                              osd_mode = 1;
                    }
                    D_INFO( "sample2: using OSD mode \"%s\".\n",
                              osd_mode ? "window" : "texture" );
               } else
               if (!strcmp( option, "-no-logo" )) {
                    use_logo = false;
                    D_INFO( "sample2: logo disabled.\n" );
               }
          } else 
          if (i == argc-1) {
               mrl = option;
          }
     }

     if (!mrl || !*mrl)
          usage();
}


static void
create_window( IDirectFBDisplayLayer  *layer,
               IDirectFBEventBuffer   *buffer,
               int                     width,
               int                     height,
               IDirectFBWindow       **ret_window )
{
     IDirectFBWindow      *window = NULL;
     DFBWindowDescription   dsc;
     
     dsc.flags  = DWDESC_POSX | DWDESC_POSY | DWDESC_WIDTH | DWDESC_HEIGHT;
     dsc.posx   = 0;
     dsc.posy   = 0;
     dsc.width  = width;
     dsc.height = height;

     if (format) {
          dsc.flags       |= DWDESC_PIXELFORMAT;
          dsc.pixelformat  = format;
     }

     DFBCHECK( layer->CreateWindow( layer, &dsc, &window ) );
     window->SetOpacity( window, 0xff );
     window->AttachEventBuffer( window, buffer );
     
     *ret_window = window;
}


static void
create_logo( IDirectFB         *dfb,
             IDirectFBSurface  *destination,
             IDirectFBSurface **ret_logo )
{
     IDirectFBSurface             *logo = NULL;
     DFBGraphicsDeviceDescription  caps;
     DFBAccelerationMask           mask;
     DFBResult                     err;

     dfb->GetDeviceDescription( dfb, &caps );
     DFBCHECK( dfb->CreateSurface( dfb, &dfblogo_desc, &logo ) );

     /* use hardware modulation when available */
     if ((caps.blitting_flags & DSBLIT_COLORIZE) &&
         (caps.blitting_flags & DSBLIT_BLEND_ALPHACHANNEL)) {
          IDirectFBSurface      *tmp = NULL;
          DFBSurfaceDescription  dsc;

          dsc.flags       = DSDESC_CAPS | DSDESC_WIDTH | DSDESC_HEIGHT | 
                            DSDESC_PIXELFORMAT;
          dsc.caps        = DSCAPS_VIDEOONLY;
          dsc.width       = dfblogo_desc.width;
          dsc.height      = dfblogo_desc.height;
          dsc.pixelformat = dfblogo_desc.pixelformat;
               
          err = dfb->CreateSurface( dfb, &dsc, &tmp );
          if (err == DFB_OK && tmp) {
               /* check if hardware modulation for the current format */
               destination->SetBlittingFlags( destination, DSBLIT_COLORIZE |
                                                           DSBLIT_BLEND_ALPHACHANNEL );
               destination->GetAccelerationMask( destination, tmp, &mask );
               destination->SetBlittingFlags( destination, DSBLIT_NOFX );
               
               if (mask & DFXL_BLIT) {
                    tmp->Blit( tmp, logo, NULL, 0, 0 );
                    logo->Release( logo );
                    logo = tmp;
               } else
                    tmp->Release( tmp );
          } else
               DirectFBError( "couldn't create logo surface", err );
     }

     *ret_logo = logo;
}


static int 
update_visual( IDirectFB              *dfb,
               IDirectFBDisplayLayer  *layer,
               IDirectFBSurface       *destination,
               xine_video_port_t      *vo_driver,
               dfb_visual_t           *visual,
               int                     width,
               int                     height,
               IDirectFBWindow       **ret_osd_window )
{
     IDirectFBSurface             *frame      = visual->destination;
     IDirectFBSurface             *osd        = visual->subpicture;
     IDirectFBWindow              *osd_window = *ret_osd_window;
     DFBSurfaceDescription         dsc;
     DFBGraphicsDeviceDescription  caps;
     DFBAccelerationMask           mask;
     DFBResult                     err;

     if (width < 1 || height < 1)
          return 0;

     if (frame) {
          int w, h;

          frame->GetSize( frame, &w, &h );
          if (w == width && h == height)
               return 1;
          
          frame->Release( frame );
          frame = NULL;
          if (osd_mode == 0) {
               /* texture mode, reallocate surface */
               if (osd)
                    osd->Release( osd );
               osd = NULL;
          }

          D_INFO( "sample2: video size changed, reallocating buffers.\n" );
     }

     dfb->GetDeviceDescription( dfb, &caps );
     
     if (caps.acceleration_mask & DFXL_STRETCHBLIT) {
          /* create buffer for video frames */
          dsc.flags  = DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_CAPS;
          dsc.caps   = DSCAPS_VIDEOONLY;
          dsc.width  = width;
          dsc.height = height;
          
          if (format) {
               dsc.flags       |= DSDESC_PIXELFORMAT;
               dsc.pixelformat  = format;
          }

          err = dfb->CreateSurface( dfb, &dsc, &frame );
          if (err != DFB_OK) { 
               DirectFBError( "couldn't create frame surface", err );
               if (err == DFB_NOVIDEOMEMORY) {
                    /* not enough memory, disable hardware scaling */
                    frame = NULL;
                    mask  = DFXL_NONE;
               } else
                    return 0;
          } else {
               /* check if hardware StechBlit is supported for the current format */
               destination->SetBlittingFlags( destination, DSBLIT_NOFX );
               destination->GetAccelerationMask( destination, frame, &mask );
          }
               
          if (!(mask & DFXL_STRETCHBLIT)) {
               /* fall back to software scaling */
               if (frame)
                    frame->Release( frame );
               destination->AddRef( destination );
               frame = destination;
          } else
               D_INFO( "sample2: using hardware StretchBlit.\n" );
     }
     else {
          /* fall back to software scaling */
          destination->AddRef( destination );
          frame = destination;
     }

     /* check if hardware acceleration can be used for OSD */
     if (use_hwosd && (caps.blitting_flags & DSBLIT_BLEND_ALPHACHANNEL)) {
          if (osd_mode == 1) {
               /* window mode */
               if (!osd_window) {
                    DFBWindowDescription w_dsc;
                    DFBWindowOptions     opt;

                    w_dsc.flags  = DWDESC_WIDTH | DWDESC_HEIGHT | DWDESC_CAPS;
                    w_dsc.width  = dsc.width;
                    w_dsc.height = dsc.height;
                    w_dsc.caps   = DWCAPS_ALPHACHANNEL | DWCAPS_NODECORATION;

                    err = layer->CreateWindow( layer, &w_dsc, &osd_window);
                    if (err == DFB_OK) {
                        osd_window->SetOpacity( osd_window, 0x00 );
                        osd_window->GetOptions( osd_window, &opt );
                        osd_window->SetOptions( osd_window, opt | DWOP_GHOST );
                        osd_window->GetSurface( osd_window, &osd );
                        osd->Clear( osd, 0, 0, 0, 0 );
                        osd->Flip( osd, NULL, 0 );
                        osd_window->SetOpacity( osd_window, 0xff );
                    } else
                         D_ERROR( "sample2: couldn't create OSD window (%s)\n",
                                   DirectFBErrorString( err ) );
               }
          }
          else {
               /* texture mode */
               dsc.flags       |= DSDESC_CAPS | DSDESC_PIXELFORMAT;
               dsc.caps         = DSCAPS_VIDEOONLY;
               dsc.pixelformat  = DSPF_ARGB;

               err = dfb->CreateSurface( dfb, &dsc, &osd );
               if (err == DFB_OK) {
                    frame->SetBlittingFlags( frame, DSBLIT_BLEND_ALPHACHANNEL );
                    frame->GetAccelerationMask( frame, osd, &mask );
               
                    if (mask & DFXL_BLIT) {
                         D_INFO( "sample2: using hardware OSD.\n" );
                         osd->Clear( osd, 0, 0, 0, 0 );
                    } else {
                         osd->Release( osd );
                         osd = NULL;
                    }
               } else {
                    DirectFBError( "couldn't create OSD surface", err );
                    osd = NULL;
               }
          }
     }

     /* update visual */
     visual->destination = frame;
     visual->subpicture  = osd;
     if (!xine_port_send_gui_data( vo_driver,
                    XINE_GUI_SEND_SELECT_VISUAL, (void*) visual )) {
          D_ERROR( "sample2: failed to update driver's visual!!\n" );
          return 0;
     }

     *ret_osd_window = osd_window;

     return 1;
}


static void
seek( xine_stream_t *stream,
      int            step )
{
     int pos = -1;
     int try;

     if (!xine_get_stream_info( stream, XINE_STREAM_INFO_SEEKABLE ))
          return;

     for (try = 5; try--; ) {
          if (xine_get_pos_length( stream, NULL,  &pos, NULL ))
               break;
          usleep( 1000 );
     }

     if (pos >= 0) {
          pos += step;
          pos  = max( pos, 0 );
          xine_play( stream, 0, pos );
          usleep( 10 );
     }
}

static void
adjust_color( xine_stream_t           *stream,
              DFBColorAdjustmentFlags  flags,
              int                      step )
{
     int value;
     
     if (flags & DCAF_BRIGHTNESS)  {
          value = xine_get_param( stream, XINE_PARAM_VO_BRIGHTNESS ); 
          xine_set_param( stream,
                          XINE_PARAM_VO_BRIGHTNESS,
                          CLAMP( value+step, 0, 0xffff ) );
     }

     if (flags & DCAF_CONTRAST) {
          value = xine_get_param( stream, XINE_PARAM_VO_CONTRAST );
          xine_set_param( stream,
                          XINE_PARAM_VO_CONTRAST,
                          CLAMP( value+step, 0, 0xffff ) );
     }

     if (flags & DCAF_SATURATION) {
          value = xine_get_param( stream, XINE_PARAM_VO_SATURATION );
          xine_set_param( stream,
                          XINE_PARAM_VO_SATURATION,
                          CLAMP( value+step, 0, 0xffff ) );
     }

     /* HUE is not supported in video_out_dfb */
}

static void
show_status( xine_osd_t    *osd,
             xine_stream_t *stream )
{
     int   status = xine_get_status( stream );
     int   speed  = xine_get_param( stream, XINE_PARAM_SPEED );
     char *symbol = "";

     switch (status) {
          case XINE_STATUS_IDLE:
               symbol = "";
               break;
          case XINE_STATUS_STOP:
               symbol = "}";
               break;
          case XINE_STATUS_PLAY:
               symbol = (speed == XINE_SPEED_PAUSE) ? "<" : ">";
               break;
          case XINE_STATUS_QUIT:
               symbol = "{";
               break;
          default:
               break;
     }

     xine_osd_clear( osd );
     xine_osd_draw_text( osd, 0, 0, symbol, XINE_OSD_TEXT1 );
     if (osd_mode == 1) {
          /* important: the osd will be rendered at screen resolution */
          xine_osd_show_unscaled( osd, 0 );
     } else
          xine_osd_show( osd, 0 );

     /* hide in 5 seconds */
     xine_osd_hide( osd, xine_get_current_vpts( stream ) + 5*90000 );
}


static void
output_cb( void *cdata, int width, int height, double ratio,
           DFBSurfacePixelFormat format, DFBRectangle *dst )
{
     struct frame_data *data = (struct frame_data*) cdata;
     static int         old_width  = 0;
     static int         old_height = 0;
     static double      old_ratio  = 0.0;

     /* remember size (used to decide buffers reallocation) */
     data->width  = width;
     data->height = height;

     /* check if size or ratio changed */
     if (width != old_width || height != old_height || ratio != old_ratio) {
          IDirectFBWindow *window  = data->window;

          /* change window size respecting the aspect ratio */
          if (data->screen_ratio > ratio)
               window->Resize( window, (double)height*ratio, height );
          else
               window->Resize( window, width, (double)width/ratio );

          old_width  = width;
          old_height = height;
          old_ratio  = ratio;
     }
     
     dst->x = 0;
     dst->y = 0;
     data->frame->GetSize( data->frame, &dst->w, &dst->h );
}

static int
frame_cb( void *cdata )
{
	static DFBColor    color    = { 0x22, 0x33, 0xbb, 0xff };
     struct frame_data *data     = (struct frame_data*) cdata;
     IDirectFBSurface  *frame    = data->frame;
     IDirectFBSurface  *osd      = data->osd;
     IDirectFBSurface  *logo     = data->logo;
     IDirectFBSurface  *dest     = data->dest;
     int                sw, sh;
     int                dw, dh;
     
     dest->GetSize( dest, &dw, &dh );
     
     pthread_mutex_lock( &data->mutex );
     
     frame->GetSize( frame, &sw, &sh );

     /* check if we need to blit the OSD surface (texture mode) */
     if (osd && osd_mode == 0) {
          int w, h;
          osd->GetSize( osd, &w, &h );
          frame->SetBlittingFlags( frame, DSBLIT_BLEND_ALPHACHANNEL );
          if (w != sw || h != sh)
               frame->StretchBlit( frame, osd, NULL, NULL );
          else
               frame->Blit( frame, osd, NULL, 0, 0 );
     }
     
     if (frame != dest) {
          dest->SetBlittingFlags( dest, DSBLIT_NOFX );
          /* using hardware stretchblit */
          if (dw != sw || dh != sh)
               dest->StretchBlit( dest, frame, NULL, NULL );
          else
               dest->Blit( dest, frame, NULL, 0, 0 );
     }

     pthread_mutex_unlock( &data->mutex );

     /* draw progressive logo */
     if (logo) {
          DFBRectangle rect0;
          DFBRectangle rect1;

          /* compute coordinates whitin logo surface */
          rect0.y = rect1.y = 0;
          rect0.h = rect1.h = dfblogo_desc.height;
          rect0.x = 0;
          rect0.w = dfblogo_desc.width * data->progress / 100;
          rect1.x = rect0.w;
          rect1.w = dfblogo_desc.width - rect0.w;
          
          /* elapsed */
          dest->SetColor( dest, color.r, color.g, color.b, 0xff );
          dest->SetBlittingFlags( dest, DSBLIT_COLORIZE |
                                        DSBLIT_BLEND_ALPHACHANNEL );
          dest->Blit( dest, logo, &rect0, 
                      7, dh - dfblogo_desc.height - 7 );
          /* remaining */
          dest->SetBlittingFlags( dest, DSBLIT_BLEND_ALPHACHANNEL );
          dest->Blit( dest, logo, &rect1,
                      7 + rect0.w, dh - dfblogo_desc.height - 7 );
          /* rotate colors */
          color.r -= 2;
          color.g += 1;
          color.b -= 2;
     }

     dest->Flip( dest, NULL, 0 );

     return 0;
}


int
main( int argc, char **argv )
{
     IDirectFB                    *dfb        = NULL; 
     IDirectFBDisplayLayer        *layer      = NULL;
     IDirectFBEventBuffer         *buffer     = NULL;
     IDirectFBSurface             *logo       = NULL;
     IDirectFBWindow              *window     = NULL;
     IDirectFBSurface             *surface    = NULL;
     IDirectFBWindow              *osd_window = NULL;
     DFBDisplayLayerConfig         l_config;
     xine_t                       *xine       = NULL;
     xine_video_port_t            *vo         = NULL;
     xine_audio_port_t            *ao         = NULL;
     xine_stream_t                *stream     = NULL;
     xine_osd_t                   *osd        = NULL;
     dfb_visual_t                  visual;
     static struct frame_data      data;
     char                         *config;
     
     if (argc < 2)
          usage();

     DFBCHECK( DirectFBInit( &argc, &argv ) );
     
     parse_options( argc, argv );

     DFBCHECK( DirectFBCreate( &dfb ) );
     DFBCHECK( dfb->CreateEventBuffer( dfb, &buffer ) );

     DFBCHECK( dfb->GetDisplayLayer( dfb, DLID_PRIMARY, &layer ) );
     layer->GetConfiguration( layer, &l_config );
     
     xine = xine_new();
     if (!xine) {
          D_ERROR( "sample2: xine_new() failed!\n" );
          goto end;
     }

     asprintf( &config, "%s/.xine/config", xine_get_homedir() );
     if (config)
          xine_config_load( xine, config );
     
     xine_init( xine );
     /* be verbose */
     xine_engine_set_param( xine,
                            XINE_ENGINE_PARAM_VERBOSITY,
                            XINE_VERBOSITY_LOG );

     /* fill the visual */
     visual.destination  = NULL;
     visual.subpicture   = NULL;
     visual.output_cb    = output_cb;
     visual.output_cdata = (void*) &data;
     visual.frame_cb     = frame_cb;
     visual.frame_cdata  = (void*) &data;
     /* open DFB video driver */
     vo = xine_open_video_driver( xine, "DFB",
                                  XINE_VISUAL_TYPE_DFB, (void*) &visual );
     if (!vo) {
          D_ERROR( "sample2: couldn't open video driver DFB!\n" );
          goto end;
     }

     /* open audio driver (try Alsa) */
     ao = xine_open_audio_driver( xine, "alsa", NULL );
     if (!ao) {
          /* try OSS */
          ao = xine_open_audio_driver( xine, "oss", NULL );
          if (!ao) {
               D_ERROR( "sample2: couldn't find a working audio driver!\n" );
               /* disable audio */
               ao = xine_open_audio_driver( xine, "null", NULL );
               xine_engine_set_param( xine,
                                      XINE_PARAM_IGNORE_AUDIO,
                                      1 );
          }
     }

     stream = xine_stream_new( xine, ao, vo );
     if (!stream) {
          D_ERROR( "sample2: xine_stream_new() failed!\n" );
          goto end;
     }
     /* be verbose */
     xine_set_param( stream,
                     XINE_PARAM_VERBOSITY,
                     XINE_VERBOSITY_LOG );

     if (!xine_open( stream, mrl )) {
          D_ERROR( "sample2: xine_open( \"%s\" ) unsupported!\n", mrl );
          goto end;
     }
     
     /* query video size */
     data.width  = xine_get_stream_info( stream,
                                        XINE_STREAM_INFO_VIDEO_WIDTH );
     data.height = xine_get_stream_info( stream,
                                        XINE_STREAM_INFO_VIDEO_HEIGHT );
     data.width  = (data.width < 1)  ? 320 : data.width;
     data.height = (data.height < 1) ? 240 : data.height;
     
     D_INFO( "sample2: video size is %ix%i.\n", data.width, data.height );
     
     /* create window */
     create_window( layer, buffer, data.width, data.height, &window );
     window->GetSurface( window, &surface );
     surface->Clear( surface, 0, 0, 0, 0xff );
     surface->Flip( surface, NULL, 0 );
 
     if (use_logo)
          create_logo( dfb, surface, &logo );
     
     /* setup buffers and update video driver's visual */
     if (!update_visual( dfb, layer, surface, vo, &visual,
                         data.width, data.height, &osd_window )) {
          D_ERROR( "sample2: update_visual() failed!!\n" );
          goto end;
     }

     data.window       = window;
     data.frame        = visual.destination;
     data.osd          = visual.subpicture;
     data.logo         = logo;
     data.dest         = surface;
     data.progress     = 0;
     data.screen_ratio = (double) l_config.width /
                         (double) l_config.height;
     pthread_mutex_init( &data.mutex, NULL );

     /* create the osd renderer */
     osd = xine_osd_new( stream, 0, 0, 50, 50 );
     if (!osd) {
          D_ERROR( "sample2: xine_osd_new() failed!\n" );
          return false;
     }
     xine_osd_set_font( osd, "cetus", 20 );
     xine_osd_set_text_palette( osd,
                                XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT,
                                XINE_OSD_TEXT1 );
     xine_osd_set_position( osd, 15, 15 );

     /* play */
     xine_play( stream, 0, 0 );
     usleep( 500 );

     show_status( osd, stream );

     while (1) {
          static int     v_width  = 0;
          static int     v_height = 0;
          static int     adj      = 0;
          static bool    active   = false;
          static bool    moved    = false;
          static int     w_width  = 0;
          static int     w_height = 0;
          static int     startx;
          static int     starty;
          static int     endx;
          static int     endy;
          DFBWindowEvent evt;
          int            speed;

          /* query window size */
          if (!w_width || !w_height)
               window->GetSize( window, &w_width, &w_height );

          /* check if buffers need reallocation */
          if (v_width != data.width || v_height != data.height) { 
               pthread_mutex_lock( &data.mutex );
               
               /* update visual */
               if (!update_visual( dfb, layer, surface, vo, &visual,
                                   data.width, data.height, &osd_window )) {
                    D_ERROR( "sample2: update_visual() failed!!\n" );
                    pthread_mutex_unlock( &data.mutex );
                    goto end;
               }
               /* update data for frame callback */
               data.frame = visual.destination;
               data.osd   = visual.subpicture;
               
               pthread_mutex_unlock( &data.mutex );

               v_width  = data.width;
               v_height = data.height;
          }

          if (buffer->WaitForEventWithTimeout( buffer, 0, 80 ) ==  DFB_TIMEOUT) {
               int len = 0, pos = 0;
               xine_get_pos_length( stream, NULL, &pos, &len );
               data.progress = pos * 100 / (len ? : 1);
               continue;
          }

          while (buffer->GetEvent( buffer, DFB_EVENT(&evt)) == DFB_OK) {
               xine_input_data_t input = {
                    .event = {
                         .stream      = stream,
                         .data        = (void*) &input,
                         .data_length = sizeof(xine_input_data_t),
                    },
               };
                         
               switch (evt.type) {
                    case DWET_KEYDOWN:
                         switch (tolower(evt.key_symbol)) {
                              case DIKS_ESCAPE:
                              case DIKS_SMALL_Q:
                                   goto end;

                              case DIKS_ENTER:
                                   speed = xine_get_param( stream,
                                                           XINE_PARAM_SPEED );
                                   
                                   if (speed != XINE_SPEED_PAUSE)
                                        xine_set_param( stream,
                                                        XINE_PARAM_SPEED,
                                                        XINE_SPEED_PAUSE );
                                   else
                                        xine_set_param( stream,
                                                        XINE_PARAM_SPEED,
                                                        XINE_SPEED_NORMAL );
                                   show_status( osd, stream );
                                   break;

                              case DIKS_SMALL_B:
                                   adj |= DCAF_BRIGHTNESS;
                                   break;
                              case DIKS_SMALL_C:
                                   adj |= DCAF_CONTRAST;
                                   break;
                              case DIKS_SMALL_S:
                                   adj |= DCAF_SATURATION;
                                   break;

                              case DIKS_CURSOR_LEFT:
                                   if (adj)
                                        adjust_color( stream, adj, -257 );
                                   else
                                        seek( stream, -10000 );
                                   break;

                              case DIKS_CURSOR_RIGHT:
                                   if (adj)
                                        adjust_color( stream, adj, +257 );
                                   else
                                        seek( stream, +10000 );
                                   break;

                              default:
                                   break;
                         }
                         break;

                    case DWET_KEYUP:
                         switch (tolower(evt.key_symbol)) {
                              case DIKS_SMALL_B:
                                   adj &= ~DCAF_BRIGHTNESS;
                                   break;
                              case DIKS_SMALL_C:
                                   adj &= ~DCAF_CONTRAST;
                                   break;
                              case DIKS_SMALL_S:
                                   adj &= ~DCAF_SATURATION;
                                   break;
                              default:
                                   break;
                         }
                         break;
                    
                    case DWET_BUTTONDOWN:                         
                         if (evt.button == DIBI_LEFT) { 
                              layer->GetCursorPosition( layer,
                                                        &startx, &starty );
                              window->GrabPointer( window );
                              window->RaiseToTop( window );
                              active = true;
                         }
                         input.event.type = XINE_EVENT_INPUT_MOUSE_BUTTON;
                         input.button     = evt.button - DIBI_LEFT + 1;
                         input.x          = evt.x * v_width  / w_width;
                         input.y          = evt.y * v_height / w_height;
                         gettimeofday( &input.event.tv, NULL );
                         xine_event_send( stream, &input.event );
                         break;

                    case DWET_BUTTONUP:
                         if (evt.button == DIBI_LEFT && active) {
                              window->UngrabPointer( window );
                              active = false;
                              moved  = false;
                         }
                         break;

                    case DWET_MOTION:
                         if (active) {
                              moved = true;
                              endx  = evt.cx;
                              endy  = evt.cy;
                         }
                         input.event.type = XINE_EVENT_INPUT_MOUSE_MOVE;
                         input.x          = evt.x * v_width  / w_width;
                         input.y          = evt.y * v_height / w_height;
                         gettimeofday( &input.event.tv, NULL );
                         xine_event_send( stream, &input.event );
                         break;

                    case DWET_POSITION:
                         if (osd_window)
                              osd_window->MoveTo( osd_window, evt.x, evt.y );
                         break;
                         
                    case DWET_SIZE:
                         w_width  = evt.w;
                         w_height = evt.h;
                         if (osd_window)
                              osd_window->Resize( osd_window, evt.w, evt.h );
                         break;

                    case DWET_ENTER:
                    case DWET_LEAVE:
                    case DWET_GOTFOCUS:
                    case DWET_LOSTFOCUS:
                         if (osd_window)
                              osd_window->PutAtop( osd_window, window );
                         break;
                         
                    case DWET_CLOSE:
                    case DWET_DESTROYED:
                         goto end;

                    default:
                         break;
               }

               if (active && moved) {
                    window->Move( window, endx - startx, endy - starty ); 
                    startx = endx;
                    starty = endy;
                    moved  = false;
               }
          }
     }


end:
     if (osd) {
          xine_osd_hide( osd, 0 );
          xine_osd_free( osd );
     }
     
     if (stream) {
          xine_stop( stream );
          xine_close( stream );
          xine_dispose( stream );
     }

     if (ao)
          xine_close_audio_driver( xine, ao );

     if (vo)
          xine_close_video_driver( xine, vo );

     if (config) {
          xine_config_save( xine, config );
          free( config );
     }

     if (xine)
          xine_exit( xine );

     if (buffer)
          buffer->Release( buffer );

     if (data.frame)
          data.frame->Release( data.frame );

     if (data.osd)
          data.osd->Release( data.osd );

     if (surface)
          surface->Release( surface );

     if (window)
          window->Release( window );

     if (osd_window)
          osd_window->Release( osd_window );

     if (logo)
          logo->Release( logo );

     if (layer)
          layer->Release( layer );

     if (dfb)
          dfb->Release( dfb );

     pthread_mutex_destroy( &data.mutex );

     return 0;
}

